"""
This module provides the OrbitList class.
"""

from typing import List

import numpy as np

from _icet import _OrbitList
from ase import Atoms
from icet.core.local_orbit_list_generator import LocalOrbitListGenerator
from icet.core.neighbor_list import get_neighbor_lists
from icet.core.permutation_matrix import (_get_lattice_site_permutation_matrix,
                                          permutation_matrix_from_structure)
from icet.core.structure import Structure
from icet.io.logging import logger

logger = logger.getChild('orbit_list')


class OrbitList(_OrbitList):
    """
    The orbit list object handles an internal list of orbits.

    An orbit has a list of equivalent sites with the restriction
    that at least one site is in the cell of the primitive structure.

    Parameters
    ----------
    structure : Atoms
        This structure will be used to construct a primitive
        structure on which all the lattice sites in the orbits
        are based.
    cutoffs : list of float
        the i-th element of this list is the cutoff for orbits with
        order i+2
    """

    def __init__(self, structure_in, cutoffs):
        if isinstance(structure_in, Structure):
            structure_in = structure_in.to_structure()
        structure = structure_in.copy()
        structure.wrap()
        max_cutoff = np.max(cutoffs)
        # Set up a permutation matrix
        permutation_matrix, prim_structure, _ \
            = permutation_matrix_from_structure(structure, max_cutoff, find_prim=False)

        logger.info('Done getting permutation_matrix.')

        # Get a list of neighbor-lists
        neighbor_lists = get_neighbor_lists(prim_structure, cutoffs)

        logger.info('Done getting neighbor lists.')

        # Transform permutation_matrix to be in lattice site format
        pm_lattice_sites \
            = _get_lattice_site_permutation_matrix(prim_structure,
                                                   permutation_matrix,
                                                   prune=True)

        logger.info('Transformation of permutation matrix to lattice neighbor'
                    'format completed.')

        _OrbitList.__init__(self, prim_structure,
                            pm_lattice_sites, neighbor_lists)
        self.sort()
        logger.info('Finished construction of orbit list.')

    @property
    def primitive_structure(self):
        """
        Returns the primitive structure to which the lattice sites in
        the orbits are referenced to.
        """
        return self._primitive_structure.copy()

    def __str__(self):
        """String representation."""
        nice_str = 'Number of orbits: {}'.format(len(self))

        for i, orbit in enumerate(self.orbits):
            cluster_str = self.orbits[i].representative_cluster.__str__()
            nice_str += "\norbit {} - Multiplicity {} - Cluster: {}".format(
                i, len(orbit), cluster_str)
        return nice_str

    def get_supercell_orbit_list(self, structure: Atoms):
        """
        Returns an orbit list for a supercell structure.

        Parameters
        ----------
        structure
            supercell atomic structure

        Returns
        -------
        An OrbitList object
        """
        log = LocalOrbitListGenerator(self, Structure.from_atoms(structure))

        supercell_orbit_list = log.generate_full_orbit_list()

        return supercell_orbit_list

    def remove_inactive_orbits(self, allowed_species: List[List[str]]) -> None:
        """ Removes orbits with inactive sites.

        Parameters
        ----------
        allowed_species
            the list of allowed species on each site in the primitive
            structure
        """
        prim_structure = self.get_primitive_structure()
        number_of_allowed_species = [len(sym) for sym in allowed_species]
        prim_structure.set_number_of_allowed_species(number_of_allowed_species)
        self._remove_inactive_orbits(prim_structure)



from collections import OrderedDict
from typing import List

import numpy as np

from ase import Atoms
from icet import ClusterSpace
from icet.core.sublattices import Sublattices
from mchammer.calculators.base_calculator import BaseCalculator


class TargetVectorCalculator(BaseCalculator):
    """
    A ``TargetVectorCalculator`` enables evaluation of the similarity
    between a structure and a target cluster vector. Such a comparison
    can be carried out in many ways, and this implementation follows the
    measure proposed by van de Walle *et al.* in Calphad **42**, 13
    (2013) [WalTiwJon13]_. Specifically, the objective function
    :math:`Q` is calculated as

    .. math::
        Q = - \\omega L + \\sum_{\\alpha}
         \\left| \\Gamma_{\\alpha} - \\Gamma^{\\text{target}}_{\\alpha}
         \\right|.

    Here, :math:`\\Gamma_{\\alpha}` are components in the cluster vector
    and :math:`\\Gamma^\\text{target}_{\\alpha}` the corresponding
    target values. The factor :math:`\\omega` is the radius of the
    largest pair cluster such that all clusters with the same or smaller
    radii have :math:`\\Gamma_{\\alpha} -
    \\Gamma^\\text{target}_{\\alpha} = 0`.

    Parameters
    ----------
    structure
        structure for which to set up calculator
    cluster_space
        cluster space from which to build calculator
    target_vector
        vector to which any vector will be compared
    weights
        weighting of each component in cluster vector
        comparison, by default 1.0 for all components
    optimality_weight
        factor :math:`L`, a high value of which effectively
        favors a complete series of optimal cluster correlations
        for the smallest pairs (see above)
    optimality_tol
        tolerance for determining whether a perfect match
        has been achieved (used in conjunction with :math:`L`)
    name
        human-readable identifier for this calculator
    """

    def __init__(self, structure: Atoms, cluster_space: ClusterSpace,
                 target_vector: List[float],
                 weights: List[float] = None,
                 optimality_weight: float = 1.0,
                 optimality_tol: float = 1e-5,
                 name: str = 'Target vector calculator') -> None:
        super().__init__(structure=structure, name=name)

        if len(target_vector) != len(cluster_space):
            raise ValueError('Cluster space and target vector '
                             'must have the same length')
        self.cluster_space = cluster_space
        self.target_vector = target_vector

        if weights is None:
            weights = np.array([1.0] * len(cluster_space))
        else:
            if len(weights) != len(cluster_space):
                raise ValueError('Cluster space and weights '
                                 'must have the same length')
        self.weights = np.array(weights)

        if optimality_weight is not None:
            self.optimality_weight = optimality_weight
            self.optimality_tol = optimality_tol
            self.orbit_data = self.cluster_space.orbit_data
        else:
            self.optimality_weight = None
            self.optimality_tol = None
            self.orbit_data = None

        self._cluster_space = cluster_space

    def calculate_total(self, occupations: List[int]) -> float:
        """
        Calculates and returns the similarity value :math:`Q`
        of the current configuration.

        Parameters
        ----------
        occupations
            the entire occupation vector (i.e. list of atomic species)
        """
        self.structure.set_atomic_numbers(occupations)
        cv = self.cluster_space.get_cluster_vector(self.structure)
        return compare_cluster_vectors(cv, self.target_vector,
                                       self.orbit_data,
                                       weights=self.weights,
                                       optimality_weight=self.optimality_weight,
                                       tol=self.optimality_tol)

    def calculate_local_contribution(self, local_indices, occupations: List[int]) -> float:
        """
        Not yet implemented, forwards calculation to
        calculate_total.
        """
        return self.calculate_total(occupations)

    @property
    def sublattices(self) -> Sublattices:
        """Sublattices of the calculators structure."""
        sl = self.cluster_space.get_sublattices(self.structure)
        return sl


def compare_cluster_vectors(cv_1: np.ndarray, cv_2: np.ndarray,
                            orbit_data: OrderedDict,
                            weights: List[float] = None,
                            optimality_weight: float = 1.0,
                            tol: float = 1e-5) -> float:
    """
    Calculate a quantity that measures similarity between two cluster
    vecors.

    Parameters
    ----------
    cv_1
        cluster vector 1
    cv_2
        cluster vector 2
    orbit_data
        orbit data as obtained by ``ClusterSpace.orbit_data``
    weights
        Weight assigned to each cluster vector element
    optimality_weight
        quantity :math:`L` in [WalTiwJon13]_
        (see :class:`mchammer.calculators.TargetVectorCalculator`)
    tol
        numerical tolerance for determining whether two elements are
        exactly equal
    """
    if weights is None:
        weights = np.ones(len(cv_1))
    diff = abs(cv_1 - cv_2)
    score = np.dot(diff, weights)
    if optimality_weight:
        longest_optimal_radius = 0
        for orbit_index, d in enumerate(diff):
            orbit = orbit_data[orbit_index]
            if orbit['order'] != 2:
                continue
            if d < tol:
                longest_optimal_radius = orbit['radius']
            else:
                break
        score -= optimality_weight * longest_optimal_radius
    return score



from typing import List, Union

import numpy as np

from _icet import _ClusterExpansionCalculator
from icet.io.logging import logger
from ase import Atoms
from icet import ClusterExpansion, Structure
from icet.core.sublattices import Sublattices
from mchammer.calculators.base_calculator import BaseCalculator


class ClusterExpansionCalculator(BaseCalculator):
    """A ClusterExpansionCalculator object enables the efficient
    calculation of properties described by a cluster expansion. It is
    specific for a particular (supercell) structure and commonly
    employed when setting up a Monte Carlo simulation, see
    :ref:`ensembles`.

    Cluster expansions, e.g., of the energy, typically yield property
    values *per site*. When running a Monte Carlo simulation one,
    however, considers changes in the *total* energy of the
    system. The default behavior is therefore to multiply the output
    of the cluster expansion by the number of sites. This behavior can
    be changed via the ``scaling`` keyword parameter.

    Parameters
    ----------
    structure : ase.Atoms
        structure for which to set up the calculator
    cluster_expansion : ClusterExpansion
        cluster expansion from which to build calculator
    name
        human-readable identifier for this calculator
    scaling
        scaling factor applied to the property value predicted by the
        cluster expansion
    use_local_energy_calculator
        evaluate energy changes using only the local environment; this method
        is generally *much* faster; unless you know what you are doing do *not*
        set this option to `False`
    """

    def __init__(self,
                 structure: Atoms, cluster_expansion: ClusterExpansion,
                 name: str = 'Cluster Expansion Calculator',
                 scaling: Union[float, int] = None,
                 use_local_energy_calculator: bool = True) -> None:
        super().__init__(structure=structure, name=name)

        structure_cpy = structure.copy()
        cluster_expansion.prune()

        if cluster_expansion._cluster_space.is_supercell_self_correlated(structure):
            logger.warning('The ClusterExpansionCalculator self-interacts, '
                           'which may lead to erroneous results. To avoid '
                           'self-interaction, use a larger supercell or a '
                           'cluster space with shorter cutoffs.')

        self.use_local_energy_calculator = use_local_energy_calculator
        if self.use_local_energy_calculator:
            self.cpp_calc = _ClusterExpansionCalculator(
                cluster_expansion.get_cluster_space_copy(),
                Structure.from_atoms(structure_cpy))

        self._cluster_expansion = cluster_expansion
        if scaling is None:
            self._property_scaling = len(structure)
        else:
            self._property_scaling = scaling

    @property
    def cluster_expansion(self) -> ClusterExpansion:
        """ cluster expansion from which calculator was constructed """
        return self._cluster_expansion

    def calculate_total(self, *, occupations: List[int]) -> float:
        """
        Calculates and returns the total property value of the current
        configuration.

        Parameters
        ----------
        occupations
            the entire occupation vector (i.e. list of atomic species)
        """
        self.structure.set_atomic_numbers(occupations)
        return self.cluster_expansion.predict(self.structure) * \
            self._property_scaling

    def calculate_local_contribution(self, *, local_indices: List[int],
                                     occupations: List[int]) -> float:
        """
        Calculates and returns the sum of the contributions to the property
        due to the sites specified in `local_indices`

        Parameters
        ----------
        local_indices
            sites over which to sum up the local contribution
        occupations
            entire occupation vector
        """
        if not self.use_local_energy_calculator:
            return self.calculate_total(occupations=occupations)

        self.structure.set_atomic_numbers(occupations)

        local_contribution = 0
        exclude_indices = []  # type: List[int]

        for index in local_indices:
            try:
                local_contribution += self._calculate_local_contribution(
                    index, exclude_indices=exclude_indices)
            except Exception as e:
                msg = "caugh exception {}. Try setting flag ".format(e)
                msg += "`use_local_energy_calculator to False` in init"
                raise RuntimeError(msg)

            exclude_indices.append(index)

        return local_contribution * self._property_scaling

    def _calculate_local_contribution(self, index: int, exclude_indices: List[int] = []):
        """
        Internal method to calculate the local contribution for one
        index.

        Parameters
        ----------
        index : int
            lattice index
        exclude_indices
            previously calculated indices, these indices will
            be ignored in order to avoid double counting bonds

        """
        local_cv = self.cpp_calc.get_local_cluster_vector(
            self.structure.get_atomic_numbers(), index, exclude_indices)
        return np.dot(local_cv, self.cluster_expansion.parameters)

    @property
    def sublattices(self) -> Sublattices:
        """Sublattices of the calculators structure."""
        sl = self.cluster_expansion._cluster_space.get_sublattices(self.structure)
        return sl



"""Definition of the Wang-Landau data container class."""

from warnings import warn
from collections import Counter, OrderedDict
from typing import Dict, List, Tuple, Union

import numpy as np

from ase.units import kB
from pandas import DataFrame, concat as pd_concat

from icet import ClusterSpace
from .base_data_container import BaseDataContainer


class WangLandauDataContainer(BaseDataContainer):
    """
    Data container for storing information concerned with :ref:`Wang-Landau
    simulation <wang_landau_ensemble>` performed with mchammer.

    Parameters
    ----------
    structure : ase.Atoms
        reference atomic structure associated with the data container

    ensemble_parameters : dict
        parameters associated with the underlying ensemble

    metadata : dict
        metadata associated with the data container
    """

    def _update_last_state(self,
                           last_step: int,
                           occupations: List[int],
                           accepted_trials: int,
                           random_state: tuple,
                           fill_factor: float,
                           fill_factor_history: Dict[int, float],
                           histogram=Dict[int, int],
                           entropy=Dict[int, float]):
        """Updates last state of the Wang-Landau simulation.

        Parameters
        ----------
        last_step
            last trial step
        occupations
            occupation vector observed during the last trial step
        accepted_trial
            number of current accepted trial steps
        random_state
            tuple representing the last state of the random generator
        fill_factor
            fill factor of Wang-Landau algorithm
        fill_factor_history
            evolution of the fill factor of Wang-Landau algorithm (key=MC
            trial step, value=fill factor)
        histogram
            histogram of states visited during Wang-Landau simulation
        entropy
            (relative) entropy accumulated during Wang-Landau simulation
        """
        super()._update_last_state(
            last_step=last_step,
            occupations=occupations,
            accepted_trials=accepted_trials,
            random_state=random_state)
        self._last_state['fill_factor'] = fill_factor
        self._last_state['fill_factor_history'] = fill_factor_history
        self._last_state['histogram'] = histogram
        self._last_state['entropy'] = entropy

    @property
    def fill_factor(self) -> float:
        """ final value of the fill factor in the Wang-Landau algorithm """
        return self._last_state['fill_factor']

    @property
    def fill_factor_history(self) -> DataFrame:
        """ evolution of the fill factor in the Wang-Landau algorithm """
        return DataFrame({'mctrial': list(self._last_state['fill_factor_history'].keys()),
                          'fill_factor': list(self._last_state['fill_factor_history'].values())})

    def get_entropy(self) -> DataFrame:
        """Returns the (relative) entropy from this data container accumulated
        during a :ref:`Wang-Landau simulation <wang_landau_ensemble>`. Returns
        ``None`` if the data container does not contain the required information.
        """

        if 'entropy' not in self._last_state:
            return None

        # compile entropy into DataFrame
        entropy = self._last_state['entropy']
        energy_spacing = self.ensemble_parameters['energy_spacing']
        df = DataFrame(data={'energy': energy_spacing * np.array(list(entropy.keys())),
                             'entropy': np.array(list(entropy.values()))},
                       index=list(entropy.keys()))
        # shift entropy for numerical stability
        df['entropy'] -= np.min(df['entropy'])

        return df

    def get_histogram(self) -> DataFrame:
        """Returns the histogram from this data container accumulated since the
        last update of the fill factor. Returns ``None`` if the data container
        does not contain the required information.
        """

        if 'histogram' not in self._last_state:
            return None

        # compile histogram into DataFrame
        histogram = self._last_state['histogram']
        energy_spacing = self.ensemble_parameters['energy_spacing']
        df = DataFrame(data={'energy': energy_spacing * np.array(list(histogram.keys())),
                             'histogram': np.array(list(histogram.values()))},
                       index=list(histogram.keys()))

        return df


def get_density_of_states_wl(dcs: Union[BaseDataContainer, dict]) -> Tuple[DataFrame, dict]:
    """Returns a pandas DataFrame with the total density of states from a
    :ref:`Wang-Landau simulation <wang_landau_ensemble>`. If a dict of data
    containers is provided the function also returns a dictionary that
    contains the standard deviation between the entropy of neighboring data
    containers in the overlap region. These errors should be small compared to
    the variation of the entropy across each bin.

    The function can handle both a single data container and a dict thereof.
    In the latter case the data containers must cover a contiguous energy
    range and must at least partially overlap.

    Parameters
    ----------
    dcs
        data container(s), from which to extract the density of states

    Raises
    ------
    ValueError
        if multiple data containers are provided and there are inconsistencies
        with regard to basic simulation parameters such as system size or
        energy spacing
    ValueError
        if multiple data containers are provided and there is at least
        one energy region without overlap
    """

    # preparations
    if hasattr(dcs, 'get_entropy'):
        # fetch raw entropy data from data container
        df = dcs.get_entropy()
        errors = None
        if len(dcs.fill_factor_history) == 0 or dcs.fill_factor > 1e-4:
            warn('The data container appears to contain data from an'
                 ' underconverged Wang-Landau simulation.')

    elif isinstance(dcs, dict) and isinstance(dcs[next(iter(dcs))], BaseDataContainer):
        # minimal consistency checks
        tags = list(dcs.keys())
        tagref = tags[0]
        dcref = dcs[tagref]
        for tag in tags:
            dc = dcs[tag]
            if len(dc.structure) != len(dcref.structure):
                raise ValueError('Number of atoms differs between data containers ({}: {}, {}: {})'
                                 .format(tagref, dcref.ensemble_parameters['n_atoms'],
                                         tag, dc.ensemble_parameters['n_atoms']))
            for param in ['energy_spacing', 'trial_move']:
                if dc.ensemble_parameters[param] != dcref.ensemble_parameters[param]:
                    raise ValueError('{} differs between data containers ({}: {}, {}: {})'
                                     .format(param,
                                             tagref, dcref.ensemble_parameters['n_atoms'],
                                             tag, dc.ensemble_parameters['n_atoms']))
                if len(dc.fill_factor_history) == 0 or dc.fill_factor > 1e-4:
                    warn('Data container {} appears to contain data from an'
                         ' underconverged Wang-Landau simulation.'.format(tag))

        # fetch raw entropy data from data containers
        entropies = {}
        for tag, dc in dcs.items():
            entropies[tag] = dc.get_entropy()

        # sort entropies by energy
        entropies = OrderedDict(sorted(entropies.items(), key=lambda row: row[1].energy.iloc[0]))

        # line up entropy data
        errors = {}
        tags = list(entropies.keys())
        for tag1, tag2 in zip(tags[:-1], tags[1:]):
            df1 = entropies[tag1]
            df2 = entropies[tag2]
            left_lim = np.min(df2.energy)
            right_lim = np.max(df1.energy)
            if left_lim >= right_lim:
                raise ValueError('No overlap in the energy range {}...{}.\n'
                                 .format(right_lim, left_lim) +
                                 ' The closest data containers have tags "{}" and "{}".'
                                 .format(tag1, tag2))
            df1_ = df1[(df1.energy >= left_lim) & (df1.energy <= right_lim)]
            df2_ = df2[(df2.energy >= left_lim) & (df2.energy <= right_lim)]
            offset = np.average(df2_.entropy - df1_.entropy)
            errors['{}-{}'.format(tag1, tag2)] = np.std(df2_.entropy - df1_.entropy)
            entropies[tag2].entropy = entropies[tag2].entropy - offset

        # compile entropy over the entire energy range
        data = {}
        indices = {}
        counts = Counter()
        for df in entropies.values():
            for index, en, ent in zip(df.index, df.energy, df.entropy):
                data[en] = data.get(en, 0) + ent
                counts[en] += 1
                indices[en] = index
        for en in data:
            data[en] = data[en] / counts[en]

        # center entropy to prevent possible numerical issues
        entmin = np.min(list(data.values()))
        df = DataFrame(data={'energy': np.array(list(data.keys())),
                             'entropy': np.array(np.array(list(data.values()))) - entmin},
                       index=list(indices.values()))
    else:
        raise TypeError('dcs ({}) must be a data container with entropy data'
                        ' or be a list of data containers'
                        .format(type(dcs)))

    # density of states
    df['density'] = np.exp(df.entropy) / np.sum(np.exp(df.entropy))

    return df, errors


def get_average_observables_wl(dcs: Union[BaseDataContainer, dict],
                               temperatures: List[float],
                               observables: List[str] = None,
                               boltzmann_constant: float = kB) -> DataFrame:
    """Returns the average and the standard deviation of the energy from a
    :ref:`Wang-Landau simulation <wang_landau_ensemble>` for the temperatures
    specified. If the ``observables`` keyword argument is specified
    the function will also return the mean and standard deviation of the
    specified observables.

    Parameters
    ----------
    dcs
        data container(s), from which to extract density of states
        as well as observables
    temperatures
        temperatures, at which to compute the averages
    observables
        observables, for which to compute averages; the observables
        must refer to fields in the data container
    boltzmann_constant
        Boltzmann constant :math:`k_B` in appropriate
        units, i.e. units that are consistent
        with the underlying cluster expansion
        and the temperature units [default: eV/K]

    Raises
    ------
    ValueError
        if the data container(s) do(es) not contain entropy data
        from Wang-Landau simulation
    ValueError
        if data container(s) do(es) not contain requested observable
    """

    def check_observables(dc: BaseDataContainer, observables: List[str]) -> None:
        """ Helper function that checks that observables are available in data frame. """
        if observables is None:
            return
        for obs in observables:
            if obs not in dc.data.columns:
                raise ValueError('Observable ({}) not in data container.\n'
                                 'Available observables: {}'.format(obs, dc.data.columns))

    # preparation of observables
    columns_to_keep = ['potential', 'density']
    if observables is not None:
        columns_to_keep.extend(observables)

    # check that observables are available in data container
    # and prepare comprehensive data frame with relevant information
    if hasattr(dcs, 'get_entropy'):
        check_observables(dcs, observables)
        df_combined = dcs.data.filter(columns_to_keep)
        dcref = dcs
    elif isinstance(dcs, dict):
        for dc in dcs.values():
            check_observables(dc, observables)
        df_combined = pd_concat([dc.data for dc in dcs.values()],
                                ignore_index=True).filter(columns_to_keep)
        dcref = list(dcs.values())[0]
    else:
        raise TypeError('dcs ({}) must be a data container with entropy data'
                        ' or be a list of data containers'
                        .format(type(dcs)))

    # fetch entropy and density of states from data container(s)
    df_density, _ = get_density_of_states_wl(dcs)

    # compute density for each row in data container if observable averages
    # are to be computed
    if observables is not None:
        energy_spacing = dcref.ensemble_parameters['energy_spacing']
        # NOTE: we rely on the indices of the df_density DataFrame to
        # correspond to the energy scale! This is expected to be handled in
        # the get_density_of_states function.
        bins = list(np.array(np.round(df_combined.potential / energy_spacing), dtype=int))
        data_density = [dens / bins.count(k) for k, dens in df_density.density[bins].items()]

    enref = np.min(df_density.energy)
    averages = []
    n_atoms = dcref.ensemble_parameters['n_atoms']
    for temperature in temperatures:

        # mean and standard deviation of energy
        boltz = np.exp(- (df_density.energy - enref) / temperature / boltzmann_constant)
        sumint = np.sum(df_density.density * boltz)
        en_mean = np.sum(df_density.energy / n_atoms * df_density.density * boltz) / sumint
        en_std = np.sum((df_density.energy / n_atoms) ** 2 * df_density.density * boltz) / sumint
        en_std = np.sqrt(en_std - en_mean ** 2)
        record = {'temperature': temperature,
                  'potential_mean': en_mean,
                  'potential_std': en_std}

        # mean and standard deviation of other observables
        if observables is not None:
            boltz = np.exp(- (df_combined.potential - enref) / temperature / boltzmann_constant)
            sumint = np.sum(data_density * boltz)
            for obs in observables:
                obs_mean = np.sum(data_density * boltz * df_combined[obs]) / sumint
                obs_std = np.sum(data_density * boltz * df_combined[obs] ** 2) / sumint
                obs_std = np.sqrt(obs_std - obs_mean ** 2)
                record['{}_mean'.format(obs)] = obs_mean
                record['{}_std'.format(obs)] = obs_std

        averages.append(record)

    return DataFrame.from_dict(averages)


def get_average_cluster_vectors_wl(dcs: Union[BaseDataContainer, dict],
                                   cluster_space: ClusterSpace,
                                   temperatures: List[float],
                                   boltzmann_constant: float = kB) -> DataFrame:
    """Returns the average cluster vectors from a :ref:`Wang-Landau simulation
    <wang_landau_ensemble>` for the temperatures specified.

    Parameters
    ----------
    dcs
        data container(s), from which to extract density of states
        as well as observables
    cluster_space
        cluster space to use for calculation of cluster vectors
    temperatures
        temperatures, at which to compute the averages
    boltzmann_constant
        Boltzmann constant :math:`k_B` in appropriate
        units, i.e. units that are consistent
        with the underlying cluster expansion
        and the temperature units [default: eV/K]
    """

    # fetch potential and structures
    if hasattr(dcs, 'get_entropy'):
        potential, trajectory = dcs.get_data('potential', 'trajectory')
        energy_spacing = dcs.ensemble_parameters['energy_spacing']
    elif isinstance(dcs, dict):
        potential, trajectory = [], []
        for dc in dcs.values():
            p, t = dc.get_data('potential', 'trajectory')
            potential.extend(p)
            trajectory.extend(t)
        energy_spacing = list(dcs.values())[0].ensemble_parameters['energy_spacing']
        potential = np.array(potential)
    else:
        raise TypeError('dcs ({}) must be a data container with entropy data'
                        ' or be a list of data containers'
                        .format(type(dcs)))

    # fetch entropy and density of states from data container(s)
    df_density, _ = get_density_of_states_wl(dcs)

    # compute weighted density and cluster vector for each bin in energy
    # range; the weighted density is the total density divided by the number
    # of structures that fall in the respective bin
    # NOTE: the following code relies on the indices of the df_density
    # DataFrame to correspond to the energy scale. This is expected to be
    # handled in the get_density_of_states function.
    cvs = []
    weighted_density = []
    bins = list(np.array(np.round(potential / energy_spacing), dtype=int))
    for k, structure in zip(bins, trajectory):
        cvs.append(cluster_space.get_cluster_vector(structure))
        weighted_density.append(df_density.density[k] / bins.count(k))

    # compute mean and standard deviation (std) of temperature weighted
    # cluster vector
    averages = []
    enref = np.min(potential)
    for temperature in temperatures:
        boltz = np.exp(- (potential - enref) / temperature / boltzmann_constant)
        sumint = np.sum(weighted_density * boltz)
        cv_mean = np.array([np.sum(weighted_density * boltz * cv) / sumint
                            for cv in np.transpose(cvs)])
        cv_std = np.array([np.sum(weighted_density * boltz * cv ** 2) / sumint
                           for cv in np.transpose(cvs)])
        cv_std = np.sqrt(cv_std - cv_mean ** 2)
        record = {'temperature': temperature,
                  'cv_mean': cv_mean,
                  'cv_std': cv_std}
        averages.append(record)

    return DataFrame.from_dict(averages)



from .cluster_expansion_calculator import ClusterExpansionCalculator
from .target_vector_calculator import (TargetVectorCalculator,
                                       compare_cluster_vectors)

__all__ = ['ClusterExpansionCalculator',
           'TargetVectorCalculator',
           'compare_cluster_vectors']



"""
Handling of Smith Normal Form matrices
"""

import numpy as np


class SmithNormalForm(object):
    """
    Smith Normal Form matrix.
    """

    def __init__(self, H):
        self.compute_snf(H)
        self.S = tuple([self.S_matrix[i, i] for i in range(3)])
        self.ncells = self.S[0] * self.S[1] * self.S[2]
        self.group_order = None
        self.hnfs = []

        # Help list for permuting labelings
        blocks = [self.ncells // self.S[0]]
        for i in range(1, 3):
            blocks.append(blocks[-1] // self.S[i])
        self.blocks = blocks

    def compute_snf(self, H, tol=1e-3):
        """
        Compute Smith Normal Form for 3x3 matrix. Note that H = L*S*R.

        Parameters
        ----------
        H : ndarray
            3x3 matrix
        """
        A = H.copy()
        L = np.eye(3, dtype=int)
        R = np.eye(3, dtype=int)
        while True:
            # Clear upper row and leftmost column in such
            # a way that greatest common denominator ends
            # up in A[0, 0], in a standard Smith Normal Form way
            # (Euclidean algorithm for finding greatest common divisor)
            while sorted(A[0])[1] != 0 or sorted(A[:, 0])[1] != 0:
                A, R = _gcd_reduce_row(A, R, 0)
                A, L = _gcd_reduce_column(A, L, 0)

            # Do the same thing for lower 2x2 matrix
            while sorted(A[1, 1:])[0] != 0 or sorted(A[1:, 1])[0] != 0:
                A, R = _gcd_reduce_row(A, R, 1)
                A, L = _gcd_reduce_column(A, L, 1)

            # If last diagonal entry is negative,
            # make it positive
            if A[2, 2] < 0:
                A[2, 2] = -A[2, 2]
                L[2] = -L[2]

            # Check that the diagonal entry i,i divides
            # diagonal entry i+1, i+1. Otherwise,
            # add row i+1 to i and start over.
            if A[2, 2] % A[1, 1] != 0:
                A[1] = A[1] + A[2]
                L[1] = L[1] + L[2]
            elif A[1, 1] % A[0, 0] != 0:
                A[0] = A[0] + A[1]
                L[0] = L[0] + L[1]
            else:
                break
        assert (abs(np.dot(np.dot(L, H), R) - A) < tol).all()
        self.S_matrix = A
        self.L = L

    def add_hnf(self, hnf):
        """Add HNF to SNF.

        Paramaters
        ----------
        hnf : HermiteNormalForm object
        """
        self.hnfs.append(hnf)

    def set_group_order(self):
        """
        Set group representation of an SNF matrix (the G matrix in HarFor08).
        """
        group_order = []
        for i in range(self.S[0]):
            for j in range(self.S[1]):
                for k in range(self.S[2]):
                    group_order.append([i, j, k])
        self.group_order = group_order


def _switch_rows(A, i, j):
    """
    Switch rows in matrix.

    Parameters
    ---------
    A : ndarray
        Matrix in which rows will be swapped.
    i : int
        Index of row 1 to be swapped.
    j : int
        Index of row 2 to be swapped.

    Returns
    -------
    ndarray
        Matrix with swapped rows.
    """
    row = A[j].copy()
    A[j] = A[i]
    A[i] = row
    return A


def _switch_columns(A, i, j):
    """
    Switch columns in matrix.

    Parameters
    ---------
    A : ndarray
        Matrix in which columns will be swapped.
    i : int
        Index of column 1 to be swapped.
    j : int
        Index of column 2 to be swapped.

    Returns
    -------
    ndarray
        Matrix with swapped columns.
    """
    col = A[:, j].copy()
    A[:, j] = A[:, i]
    A[:, i] = col
    return A


def _gcd_reduce_row(A, R, i):
    """
    Use column operations to make A[i, i] the greatest common
    denominator of the elements in row i and the other elements
    zero.

    Parameters
    ----------
    A : ndarray
        Matrix whose row is to be cleared.
    R : ndarray
        Matrix that should be subject to the same operations.
    i : int
        Index of row to be treated.

    Returns
    -------
    ndarray
        Treated matrix A.
    ndarray
        Matrix that has been subject to the same operations.
    """
    for j in range(i, 3):
        if A[i, j] < 0:
            A[:, j] = -1 * A[:, j]
            R[:, j] = -1 * R[:, j]
    while np.sort(A[i, i:])[1 - i] > 0:
        max_index = np.argmax(A[i, i:]) + i
        min_index = np.argmin(A[i, i:]) + i
        if max_index == min_index:
            max_index += 1
        if A[i, min_index] == 0 and i == 0:
            if np.sort(A[i])[1] > 0:
                min_index += 1
                min_index = min_index % 3
                if min_index == max_index:
                    min_index += 1
                    min_index = min_index % 3
            if A[i, min_index] == A[i, max_index]:
                tmp = min_index
                min_index = max_index
                max_index = tmp
        A[:, max_index] = A[:, max_index] - A[:, min_index]
        R[:, max_index] = R[:, max_index] - R[:, min_index]
    max_index = np.argmax(A[i])
    A = _switch_columns(A, i, max_index)
    R = _switch_columns(R, i, max_index)
    return A, R


def _gcd_reduce_column(A, L, j):
    """
    Use row operations to make A[i, i] the greatest common
    denominator of the elements in column i and the other elements
    zero.

    Parameters
    ----------
    A : ndarray
        Matrix whose column is to be cleared.
    L : ndarray
        Matrix that should be subject to the same operations.
    i : int
        Index of column to be treated.

    Returns
    -------
    ndarray
        Treated matrix A.
    ndarray
        Matrix that has been subject to the same operations.
    """
    for i in range(j, 3):
        if A[i, j] < 0:
            A[i] = -1 * A[i]
            L[i] = -1 * L[i]
    while np.sort(A[j:, j])[1 - j] > 0:
        max_index = np.argmax(A[j:, j]) + j
        min_index = np.argmin(A[j:, j]) + j
        if max_index == min_index:
            max_index += 1
        if A[min_index, j] == 0 and j == 0:
            if np.sort(A[:, j])[1] > 0:
                min_index += 1
                min_index = min_index % 3
                if min_index == max_index:
                    min_index += 1
                    min_index = min_index % 3
            if A[min_index, j] == A[max_index, j]:
                tmp = min_index
                min_index = max_index
                max_index = tmp
        A[max_index] = A[max_index] - A[min_index]
        L[max_index] = L[max_index] - L[min_index]
    max_index = np.argmax(A[:, j])
    A = _switch_rows(A, j, max_index)
    L = _switch_rows(L, j, max_index)
    return A, L


def get_unique_snfs(hnfs):
    """
    For a list of Hermite Normal Forms, obtain the set of unique Smith Normal
    Forms.

    Parameters
    ----------
    hnfs : list of HermiteNormalForm objects

    Returns
    -------
    list of SmithNormalForm objects
        The unique Smith Normal Form matrices.
    """
    snfs = []
    for hnf in hnfs:
        # Check whether the snf is new or already encountered
        snf = hnf.snf
        snf_is_new = True
        for snf_comp in snfs:
            if snf_comp.S == snf.S:
                snf_is_new = False
                snf_comp.add_hnf(hnf)
                break
        if snf_is_new:
            snf.set_group_order()
            snf.add_hnf(hnf)
            snfs.append(snf)
    return snfs



"""
This module provides the StructureContainer class.
"""

import tarfile
import tempfile

from typing import BinaryIO, List, TextIO, Tuple, Union

import numpy as np
import ase.db
from ase import Atoms

from icet import ClusterSpace
from icet.io.logging import logger
logger = logger.getChild('structure_container')


class StructureContainer:
    """
    This class serves as a container for structure objects as well as their fit
    properties and cluster vectors.

    Parameters
    ----------
    cluster_space : icet.ClusterSpace
        cluster space used for evaluating the cluster vectors

    Example
    -------
    The following snippet illustrates the initialization
    and usage of a StructureContainer object. The construction
    of a structure container is convenient for compiling the
    data needed to train a cluster expansion, i.e., a sensing
    matrix and target energies::

        from ase.build import bulk
        from icet import ClusterSpace, StructureContainer
        from icet.tools import enumerate_structures
        import numpy as np

        # create cluster space
        prim = bulk('Au')
        cs = ClusterSpace(prim, cutoffs=[7.0, 5.0],
                          chemical_symbols=[['Au', 'Pd']])

        sc = StructureContainer(cs)
        for structure in enumerate_structures(prim, range(5), ['Au', 'Pd']):
            sc.add_structure(structure,
                             properties={'my_random_energy': np.random.rand()})
        print(sc)

        sensing_matrix, target_energies = sc.get_fit_data(key='my_random_energy')
    """

    def __init__(self, cluster_space: ClusterSpace):

        if not isinstance(cluster_space, ClusterSpace):
            raise TypeError('cluster_space must be a ClusterSpace object')

        self._cluster_space = cluster_space
        self._structure_list = []

    def __len__(self) -> int:
        return len(self._structure_list)

    def __getitem__(self, ind: int):
        return self._structure_list[ind]

    def get_structure_indices(self, user_tag: str = None) -> List[int]:
        """
        Get structure indices via user_tag

        Parameters
        ----------
        user_tag
            user_tag used for selecting structures

        Returns
        -------
        list of integers
            List of structure's indices
        """
        return [i for i, s in enumerate(self) if user_tag is None or s.user_tag == user_tag]

    def _get_string_representation(self, print_threshold: int = None) -> str:
        """
        String representation of the structure container that provides an
        overview of the structures in the container.

        Parameters
        ----------
        print_threshold
            if the number of structures exceeds this number print dots

        Returns
        -------
        multi-line string
            string representation of the structure container
        """

        if len(self) == 0:
            return 'Empty StructureContainer'

        # Number of structures to print before cutting and printing dots
        if print_threshold is None or print_threshold >= len(self):
            print_threshold = len(self) + 2

        # format specifiers for fields in table
        def get_format(val):
            if isinstance(val, float):
                return '{:9.4f}'
            else:
                return '{}'

        # table headers
        default_headers = ['index', 'user_tag', 'n_atoms', 'chemical formula']
        property_headers = sorted(set(key for fs in self for key in fs.properties))
        headers = default_headers + property_headers

        # collect the table data
        str_table = []
        for i, fs in enumerate(self):
            default_data = [i, fs.user_tag, len(fs), fs.structure.get_chemical_formula()]
            property_data = [fs.properties.get(key, '') for key in property_headers]
            str_row = [get_format(d).format(d) for d in default_data+property_data]
            str_table.append(str_row)
        str_table = np.array(str_table)

        # find maximum widths for each column
        widths = []
        for i in range(str_table.shape[1]):
            data_width = max(len(val) for val in str_table[:, i])
            header_width = len(headers[i])
            widths.append(max([data_width, header_width]))

        total_width = sum(widths) + 3 * len(headers)
        row_format = ' | '.join('{:'+str(width)+'}' for width in widths)

        # Make string representation of table
        s = []
        s += ['{s:=^{n}}'.format(s=' Structure Container ', n=total_width)]
        s += ['Total number of structures: {}'.format(len(self))]
        s += [''.center(total_width, '-')]
        s += [row_format.format(*headers)]
        s += [''.center(total_width, '-')]
        for i, fs_data in enumerate(str_table, start=1):
            s += [row_format.format(*fs_data)]
            if i+1 >= print_threshold:
                s += [' ...']
                s += [row_format.format(*str_table[-1])]
                break
        s += [''.center(total_width, '=')]
        s = '\n'.join(s)

        return s

    def __repr__(self) -> str:
        """ String representation. """
        return self._get_string_representation(print_threshold=50)

    def print_overview(self, print_threshold: int = None):
        """
        Prints a list of structures in the structure container.

        Parameters
        ----------
        print_threshold
            if the number of orbits exceeds this number print dots
        """
        print(self._get_string_representation(print_threshold=print_threshold))

    def add_structure(self, structure: Atoms, user_tag: str = None,
                      properties: dict = None, allow_duplicate: bool = True,
                      sanity_check: bool = True):
        """
        Adds a structure to the structure container.

        Parameters
        ----------
        structure
            the atomic structure to be added
        user_tag
            custom user tag to label structure
        properties
            scalar properties. If properties are not specified the structure
            object will be checked for an attached ASE calculator object
            with a calculated potential energy
        allow_duplicate
             whether or not to add the structure if there already exists a
             structure with identical cluster-vector
        sanity_check
            whether or not to carry out a sanity check before adding the
            structure. This includes checking occupations and volume.
        """

        # structure must have a proper format and label
        if not isinstance(structure, Atoms):
            raise TypeError('structure must be an ASE Atoms object not {}'.format(type(structure)))

        if user_tag is not None:
            if not isinstance(user_tag, str):
                raise TypeError('user_tag must be a string not {}.'.format(type(user_tag)))

        if sanity_check:
            self._cluster_space.assert_structure_compatibility(structure)

        # check for properties in attached calculator
        if properties is None:
            properties = {}
            if structure.calc is not None:
                if not structure.calc.calculation_required(structure, ['energy']):
                    energy = structure.get_potential_energy()
                    properties['energy'] = energy / len(structure)

        # check if there exist structures with identical cluster vectors
        structure_copy = structure.copy()
        cv = self._cluster_space.get_cluster_vector(structure_copy)
        if not allow_duplicate:
            for i, fs in enumerate(self):
                if np.allclose(cv, fs.cluster_vector):
                    msg = '{} and {} have identical cluster vectors'.format(
                        user_tag if user_tag is not None else 'Input structure',
                        fs.user_tag if fs.user_tag != 'None' else 'structure')
                    msg += ' at index {}'.format(i)
                    raise ValueError(msg)

        # add structure
        structure = FitStructure(structure_copy, user_tag, cv, properties)
        self._structure_list.append(structure)

    def get_condition_number(self, structure_indices: List[int] = None,
                             key: str = 'energy') -> float:
        """ Returns the condition number for the sensing matrix.

        A very large condition number can be a sign of multicollinearity,
        read more here https://en.wikipedia.org/wiki/Condition_number

        Parameters
        ----------
        structure_indices
            list of structure indices; by default (``None``) the
            method will return all fit data available.
        key
            key of properties dictionary

        Returns
        -------
        condition number of the sensing matrix
        """
        return np.linalg.cond(self.get_fit_data(structure_indices, key)[0])

    def get_fit_data(self, structure_indices: List[int] = None,
                     key: str = 'energy') -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns fit data for all structures. The cluster vectors and
        target properties for all structures are stacked into numpy arrays.

        Parameters
        ----------
        structure_indices
            list of structure indices; by default (``None``) the
            method will return all fit data available.
        key
            key of properties dictionary

        Returns
        -------
        cluster vectors and target properties for desired structures
        """
        if structure_indices is None:
            cv_list = [s.cluster_vector for s in self._structure_list]
            prop_list = [s.properties[key] for s in self._structure_list]
        else:
            cv_list, prop_list = [], []
            for i in structure_indices:
                cv_list.append(self._structure_list[i].cluster_vector)
                prop_list.append(self._structure_list[i].properties[key])

        if cv_list is None:
            raise Exception('No available fit data for {}'
                            .format(structure_indices))

        return np.array(cv_list), np.array(prop_list)

    @property
    def cluster_space(self) -> ClusterSpace:
        """Cluster space used to calculate the cluster vectors."""
        return self._cluster_space

    @property
    def available_properties(self) -> List[str]:
        """List of the available properties."""
        return sorted(set([p for fs in self for p in fs.properties.keys()]))

    def write(self, outfile: Union[str, BinaryIO, TextIO]):
        """
        Writes structure container to a file.

        Parameters
        ----------
        outfile
            output file name or file object
        """
        # Write cluster space to tempfile
        temp_cs_file = tempfile.NamedTemporaryFile()
        self.cluster_space.write(temp_cs_file.name)

        # Write fit structures as an ASE db in tempfile
        temp_db_file = tempfile.NamedTemporaryFile()
        if self._structure_list:
            db = ase.db.connect(temp_db_file.name, type='db', append=False)

        for fit_structure in self._structure_list:
            data_dict = {'user_tag': fit_structure.user_tag,
                         'properties': fit_structure.properties,
                         'cluster_vector': fit_structure.cluster_vector}
            db.write(fit_structure.structure, data=data_dict)

        with tarfile.open(outfile, mode='w') as handle:
            handle.add(temp_db_file.name, arcname='database')
            handle.add(temp_cs_file.name, arcname='cluster_space')

    @staticmethod
    def read(infile: Union[str, BinaryIO, TextIO]):
        """
        Reads StructureContainer object from file.

        Parameters
        ----------
        infile
            file from which to read

        """
        if isinstance(infile, str):
            filename = infile
        else:
            filename = infile.name

        if not tarfile.is_tarfile(filename):
            raise TypeError('{} is not a tar file'.format(filename))

        temp_db_file = tempfile.NamedTemporaryFile()
        with tarfile.open(mode='r', name=filename) as tar_file:
            cs_file = tar_file.extractfile('cluster_space')
            temp_db_file.write(tar_file.extractfile('database').read())
            temp_db_file.seek(0)
            cluster_space = ClusterSpace.read(cs_file)
            database = ase.db.connect(temp_db_file.name, type='db')

            structure_container = StructureContainer(cluster_space)
            fit_structures = []
            for row in database.select():
                data = row.data
                fit_structure = FitStructure(row.toatoms(),
                                             user_tag=data['user_tag'],
                                             cv=data['cluster_vector'],
                                             properties=data['properties'])
                fit_structures.append(fit_structure)
            structure_container._structure_list = fit_structures

        return structure_container


class FitStructure:
    """
    This class holds a supercell along with its properties and cluster
    vector.

    Attributes
    ----------
    structure : Atoms
        supercell structure
    user_tag : str
        custom user tag
    cvs : np.ndarray
        calculated cluster vector for actual structure
    properties : dict
        dictionary of properties
    """

    def __init__(self, structure: Atoms, user_tag: str,
                 cv: np.ndarray, properties: dict = {}):
        self._structure = structure
        self._user_tag = user_tag
        self._cluster_vector = cv
        self.properties = properties

    @property
    def cluster_vector(self) -> np.ndarray:
        """calculated cluster vector"""
        return self._cluster_vector

    @property
    def structure(self) -> Atoms:
        """atomic structure"""
        return self._structure

    @property
    def user_tag(self) -> str:
        """structure label"""
        return str(self._user_tag)

    def __getattr__(self, key):
        """ Accesses properties if possible and returns value. """
        if key not in self.properties.keys():
            return super().__getattribute__(key)
        return self.properties[key]

    def __len__(self) -> int:
        """ Number of sites in the structure. """
        return len(self._structure)



from ase import Atoms
from mchammer.calculators.target_vector_calculator import TargetVectorCalculator
from .canonical_ensemble import CanonicalEnsemble
from .canonical_annealing import _cooling_exponential
import numpy as np
from typing import List
import random
from icet.io.logging import logger
logger = logger.getChild('target_cluster_vector_annealing')


class TargetClusterVectorAnnealing:
    """
    Instances of this class allow one to carry out simulated annealing
    towards a target cluster vector. Because it is impossible
    to know *a priori* which supercell shape accomodates the best
    match, this ensemble allows the annealing to be done for multiple
    :class:`ase.Atoms` objects at the same time.

    Parameters
    ----------
    structure
        atomic configurations to be used in the Monte Carlo simulation;
        also defines the initial occupation vectors
    calculators
        calculators corresponding to each :class:`Atoms <ase.Atoms>`
        object
    T_start
        artificial temperature at which annealing is started
    T_stop : float
        artificial temperature at which annealing is stopped
    random_seed : int
        seed for random number generator used in the Monte Carlo
        simulation
    """

    def __init__(self, structure: List[Atoms],
                 calculators: List[TargetVectorCalculator],
                 T_start: float = 5.0, T_stop: float = 0.001,
                 random_seed: int = None) -> None:

        if isinstance(structure, Atoms):
            raise ValueError(
                'A list of ASE Atoms (supercells) must be provided')
        if len(structure) != len(calculators):
            raise ValueError('There must be as many supercells as there '
                             'are calculators ({} != {})'.format(len(structure),
                                                                 len(calculators)))

        logger.info('Initializing target cluster vector annealing '
                    'with {} supercells'.format(len(structure)))

        # random number generator
        if random_seed is None:
            self._random_seed = random.randint(0, 1e16)
        else:
            self._random_seed = random_seed
        random.seed(a=self._random_seed)

        # Initialize an ensemble for each supercell
        sub_ensembles = []
        for ens_id, (supercell, calculator) in enumerate(zip(structure, calculators)):
            sub_ensembles.append(CanonicalEnsemble(structure=supercell,
                                                   calculator=calculator,
                                                   random_seed=random.randint(
                                                       0, 1e16),
                                                   user_tag='ensemble_{}'.format(
                                                       ens_id),
                                                   temperature=T_start,
                                                   data_container=None))
        self._sub_ensembles = sub_ensembles
        self._current_score = self._sub_ensembles[0].calculator.calculate_total(
            self._sub_ensembles[0].configuration.occupations)
        self._best_score = self._current_score
        self._best_structure = structure[0]
        self._temperature = T_start
        self._T_start = T_start
        self._T_stop = T_stop
        self._total_trials = 0
        self._accepted_trials = 0
        self._n_steps = 42

    def generate_structure(self, number_of_trial_steps: int = None) -> Atoms:
        """
        Run a structure annealing simulation.

        Parameters
        ----------
        number_of_trial_steps
            Total number of trial steps to perform. If None,
            run (on average) 3000 steps per supercell
        """
        if number_of_trial_steps is None:
            self._n_steps = 3000 * len(self._sub_ensembles)
        else:
            self._n_steps = number_of_trial_steps

        self._temperature = self._T_start
        self._total_trials = 0
        self._accepted_trials = 0
        while self.total_trials < self.n_steps:
            if self._total_trials % 1000 == 0:
                logger.info('MC step {}/{} ({} accepted trials, '
                            'temperature {:.3f}), '
                            'best score: {:.3f}'.format(self.total_trials,
                                                        self.n_steps,
                                                        self.accepted_trials,
                                                        self.temperature,
                                                        self.best_score))
            self._do_trial_step()
        return self.best_structure

    def _do_trial_step(self):
        """ Carries out one Monte Carlo trial step. """
        self._temperature = _cooling_exponential(
            self.total_trials, self.T_start, self.T_stop, self.n_steps)
        self._total_trials += 1

        # Choose a supercell
        ensemble = random.choice(self._sub_ensembles)

        # Choose two sites and swap
        sublattice_index = ensemble.get_random_sublattice_index(
            ensemble._swap_sublattice_probabilities)
        sites, species = ensemble.configuration.get_swapped_state(
            sublattice_index)

        # Update occupations so that the cluster vector (and its score)
        # can be calculated
        ensemble.configuration.update_occupations(sites, species)
        new_score = ensemble.calculator.calculate_total(
            ensemble.configuration.occupations)

        if self._acceptance_condition(new_score - self.current_score):
            self._current_score = new_score
            self._accepted_trials += 1

            # Since we are looking for the best structures we want to
            # keep track of the best one we have found as yet (the
            # current one may have a worse score)
            if self._current_score < self._best_score:
                self._best_structure = ensemble.structure
                self._best_score = self._current_score
        else:
            ensemble.configuration.update_occupations(
                sites, list(reversed(species)))

    def _acceptance_condition(self, potential_diff: float) -> bool:
        """
        Evaluates Metropolis acceptance criterion.

        Parameters
        ----------
        potential_diff
            change in the thermodynamic potential associated
            with the trial step
        """
        if potential_diff < 0:
            return True
        elif abs(self.temperature) < 1e-6:  # temperature is numerically zero
            return False
        else:
            p = np.exp(-potential_diff / self.temperature)
            return p > random.random()

    @property
    def temperature(self) -> float:
        """ Current temperature """
        return self._temperature

    @property
    def T_start(self) -> float:
        """ Starting temperature """
        return self._T_start

    @property
    def T_stop(self) -> float:
        """ Stop temperature """
        return self._T_stop

    @property
    def n_steps(self) -> int:
        """ Number of steps to carry out """
        return self._n_steps

    @property
    def total_trials(self) -> int:
        """ Number of steps carried out so far """
        return self._total_trials

    @property
    def accepted_trials(self) -> int:
        """ Number of accepted trials carried out so far """
        return self._accepted_trials

    @property
    def current_score(self) -> float:
        """ Current target vector score """
        return self._current_score

    @property
    def best_score(self) -> float:
        """ Best target vector score found so far """
        return self._best_score

    @property
    def best_structure(self) -> float:
        """ Structure most closely matching target vector so far """
        return self._best_structure



"""Definition of the canonical ensemble class."""

import numpy as np

from ase import Atoms
from ase.units import kB
from typing import List

from .. import DataContainer
from ..calculators.base_calculator import BaseCalculator
from .thermodynamic_base_ensemble import ThermodynamicBaseEnsemble


class CanonicalEnsemble(ThermodynamicBaseEnsemble):
    """Instances of this class allow one to simulate systems in the
    canonical ensemble (:math:`N_iVT`), i.e. at constant temperature
    (:math:`T`), number of atoms of each species (:math:`N_i`), and
    volume (:math:`V`).

    The probability for a particular state in the canonical ensemble is
    proportional to the well-known Boltzmann factor,

    .. math::

        \\rho_{\\text{C}} \\propto \\exp [ - E / k_B T ].

    Since the concentrations or equivalently the number of atoms of each
    species is held fixed in the canonical ensemble, a trial step must
    conserve the concentrations. This is accomplished by randomly picking two
    unlike atoms and swapping their identities. The swap is accepted with
    probability

    .. math::

        P = \\min \\{ 1, \\, \\exp [ - \\Delta E / k_B T  ] \\},

    where :math:`\\Delta E` is the change in potential energy caused by the
    swap.

    The canonical ensemble provides an ideal framework for studying the
    properties of a system at a specific concentration. Properties such as
    potential energy or phenomena such as chemical ordering at a specific
    temperature can conveniently be studied by simulating at that temperature.
    The canonical ensemble is also a convenient tool for "optimizing" a
    system, i.e., finding its lowest energy chemical ordering. In practice,
    this is usually achieved by simulated annealing, i.e. the system is
    equilibrated at a high temperature, after which the temperature is
    continuously lowered until the acceptance probability is almost zero. In a
    well-behaved system, the chemical ordering at that point corresponds to a
    low-energy structure, possibly the global minimum at that particular
    concentration.

    Parameters
    ----------
    structure : :class:`Atoms <ase.Atoms>`
        atomic configuration to be used in the Monte Carlo simulation;
        also defines the initial occupation vector
    calculator : :class:`BaseCalculator <mchammer.calculators.ClusterExpansionCalculator>`
        calculator to be used for calculating the potential changes
        that enter the evaluation of the Metropolis criterion
    temperature : float
        temperature :math:`T` in appropriate units [commonly Kelvin]
    boltzmann_constant : float
        Boltzmann constant :math:`k_B` in appropriate
        units, i.e. units that are consistent
        with the underlying cluster expansion
        and the temperature units [default: eV/K]
    user_tag : str
        human-readable tag for ensemble [default: None]
    random_seed : int
        seed for the random number generator used in the Monte Carlo
        simulation
    data_container : str
        name of file the data container associated with the ensemble
        will be written to; if the file exists it will be read, the
        data container will be appended, and the file will be
        updated/overwritten
    data_container_write_period : float
        period in units of seconds at which the data container is
        written to file; writing periodically to file provides both
        a way to examine the progress of the simulation and to back up
        the data [default: np.inf]
    ensemble_data_write_interval : int
        interval at which data is written to the data container; this
        includes for example the current value of the calculator
        (i.e. usually the energy) as well as ensembles specific fields
        such as temperature or the number of atoms of different species
    trajectory_write_interval : int
        interval at which the current occupation vector of the atomic
        configuration is written to the data container.
    sublattice_probabilities : List[float]
        probability for picking a sublattice when doing a random swap.
        This should be as long as the number of sublattices and should
        sum up to 1.


    Example
    -------
    The following snippet illustrate how to carry out a simple Monte Carlo
    simulation in the canonical ensemble. Here, the parameters of the cluster
    expansion are set to emulate a simple Ising model in order to obtain an
    example that can be run without modification. In practice, one should of
    course use a proper cluster expansion::

        from ase.build import bulk
        from icet import ClusterExpansion, ClusterSpace
        from mchammer.calculators import ClusterExpansionCalculator
        from mchammer.ensembles import CanonicalEnsemble

        # prepare cluster expansion
        # the setup emulates a second nearest-neighbor (NN) Ising model
        # (zerolet and singlet ECIs are zero; only first and second neighbor
        # pairs are included)
        prim = bulk('Au')
        cs = ClusterSpace(prim, cutoffs=[4.3], chemical_symbols=['Ag', 'Au'])
        ce = ClusterExpansion(cs, [0, 0, 0.1, -0.02])

        # prepare initial configuration
        structure = prim.repeat(3)
        for k in range(5):
            structure[k].symbol = 'Ag'

        # set up and run MC simulation
        calc = ClusterExpansionCalculator(structure, ce)
        mc = CanonicalEnsemble(structure=structure, calculator=calc,
                               temperature=600,
                               data_container='myrun_canonical.dc')
        mc.run(100)  # carry out 100 trial swaps
    """

    def __init__(self,
                 structure: Atoms,
                 calculator: BaseCalculator,
                 temperature: float,
                 user_tag: str = None,
                 boltzmann_constant: float = kB,
                 random_seed: int = None,
                 data_container: str = None,
                 data_container_write_period: float = np.inf,
                 ensemble_data_write_interval: int = None,
                 trajectory_write_interval: int = None,
                 sublattice_probabilities: List[float] = None) -> None:

        self._ensemble_parameters = dict(temperature=temperature)

        # add species count to ensemble parameters
        symbols = set([symbol for sub in calculator.sublattices
                       for symbol in sub.chemical_symbols])
        for symbol in symbols:
            key = 'n_atoms_{}'.format(symbol)
            count = structure.get_chemical_symbols().count(symbol)
            self._ensemble_parameters[key] = count

        super().__init__(
            structure=structure,
            calculator=calculator,
            user_tag=user_tag,
            random_seed=random_seed,
            data_container=data_container,
            data_container_class=DataContainer,
            data_container_write_period=data_container_write_period,
            ensemble_data_write_interval=ensemble_data_write_interval,
            trajectory_write_interval=trajectory_write_interval,
            boltzmann_constant=boltzmann_constant)

        if sublattice_probabilities is None:
            self._swap_sublattice_probabilities = self._get_swap_sublattice_probabilities()
        else:
            self._swap_sublattice_probabilities = sublattice_probabilities

    @property
    def temperature(self) -> float:
        """ Current temperature """
        return self._ensemble_parameters['temperature']

    def _do_trial_step(self):
        """ Carries out one Monte Carlo trial step. """
        sublattice_index = self.get_random_sublattice_index(self._swap_sublattice_probabilities)
        return self.do_canonical_swap(sublattice_index=sublattice_index)



"""
This module provides the ClusterSpace class.
"""

import copy
import itertools
import pickle
import tarfile
import tempfile
from collections import OrderedDict
from typing import List, Union

import numpy as np

from _icet import ClusterSpace as _ClusterSpace
from ase import Atoms
from ase.io import read as ase_read
from ase.io import write as ase_write
from icet.core.orbit_list import OrbitList
from icet.core.structure import Structure
from icet.core.sublattices import Sublattices
from icet.tools.geometry import get_occupied_primitive_structure


class ClusterSpace(_ClusterSpace):
    """This class provides functionality for generating and maintaining
    cluster spaces.

    **Note:** In icet all :class:`ase.Atoms` objects must have
    periodic boundary conditions. When carrying out cluster expansions
    for surfaces and nanoparticles it is therefore recommended to
    surround the structure with vacuum and use periodic boundary
    conditions. This can be done using e.g., :func:`ase.Atoms.center`.

    Parameters
    ----------
    structure : ase.Atoms
        atomic structure
    cutoffs : list(float)
        cutoff radii per order that define the cluster space

        Cutoffs are specified in units of Angstrom and refer to the
        longest distance between two atoms in the cluster. The first
        element refers to pairs, the second to triplets, the third
        to quadruplets, and so on. ``cutoffs=[7.0, 4.5]`` thus implies
        that all pairs distanced 7 A or less will be included,
        as well as all triplets among which the longest distance is no
        longer than 4.5 A.
    chemical_symbols : list(str) or list(list(str))
        list of chemical symbols, each of which must map to an element
        of the periodic table

        If a list of chemical symbols is provided, all sites on the
        lattice will have the same allowed occupations as the input
        list.

        If a list of list of chemical symbols is provided then the
        outer list must be the same length as the `structure` object and
        ``chemical_symbols[i]`` will correspond to the allowed species
        on lattice site ``i``.

    Examples
    --------
    The following snippets illustrate several common situations::

        from ase.build import bulk
        from ase.io import read
        from icet import ClusterSpace

        # AgPd alloy with pairs up to 7.0 A and triplets up to 4.5 A
        prim = bulk('Ag')
        cs = ClusterSpace(structure=prim, cutoffs=[7.0, 4.5],
                          chemical_symbols=[['Ag', 'Pd']])
        print(cs)

        # (Mg,Zn)O alloy on rocksalt lattice with pairs up to 8.0 A
        prim = bulk('MgO', crystalstructure='rocksalt', a=6.0)
        cs = ClusterSpace(structure=prim, cutoffs=[8.0],
                          chemical_symbols=[['Mg', 'Zn'], ['O']])
        print(cs)

        # (Ga,Al)(As,Sb) alloy with pairs, triplets, and quadruplets
        prim = bulk('GaAs', crystalstructure='zincblende', a=6.5)
        cs = ClusterSpace(structure=prim, cutoffs=[7.0, 6.0, 5.0],
                          chemical_symbols=[['Ga', 'Al'], ['As', 'Sb']])
        print(cs)

        # PdCuAu alloy with pairs and triplets
        prim = bulk('Pd')
        cs = ClusterSpace(structure=prim, cutoffs=[7.0, 5.0],
                          chemical_symbols=[['Au', 'Cu', 'Pd']])
        print(cs)

    """

    def __init__(self,
                 structure: Atoms,
                 cutoffs: List[float],
                 chemical_symbols: Union[List[str], List[List[str]]]) -> None:

        if not isinstance(structure, Atoms):
            raise TypeError('Input configuration must be an ASE Atoms object'
                            ', not type {}'.format(type(structure)))
        if not all(structure.pbc):
            raise ValueError('Input structure must have periodic boundary conditions')

        self._cutoffs = cutoffs.copy()
        self._input_structure = structure.copy()
        self._input_chemical_symbols = copy.deepcopy(chemical_symbols)
        chemical_symbols = self._get_chemical_symbols()

        self._pruning_history = []

        # set up primitive
        occupied_primitive, primitive_chemical_symbols = get_occupied_primitive_structure(
            self._input_structure, chemical_symbols)
        self._primitive_chemical_symbols = primitive_chemical_symbols
        assert len(occupied_primitive) == len(primitive_chemical_symbols)

        # set up orbit list
        self._orbit_list = OrbitList(occupied_primitive, self._cutoffs)
        self._orbit_list.remove_inactive_orbits(primitive_chemical_symbols)

        # call (base) C++ constructor
        _ClusterSpace.__init__(
            self, primitive_chemical_symbols, self._orbit_list)

    def _get_chemical_symbols(self):
        """ Returns chemical symbols using input structure and
        chemical symbols. Carries out multiple sanity checks. """

        # setup chemical symbols as List[List[str]]
        if all(isinstance(i, str) for i in self._input_chemical_symbols):
            chemical_symbols = [
                self._input_chemical_symbols] * len(self._input_structure)
        elif not all(isinstance(i, list) for i in self._input_chemical_symbols):
            raise TypeError("chemical_symbols must be List[str] or List[List[str]], not {}".format(
                type(self._input_chemical_symbols)))
        elif len(self._input_chemical_symbols) != len(self._input_structure):
            msg = 'chemical_symbols must have same length as structure. '
            msg += 'len(chemical_symbols) = {}, len(structure)= {}'.format(
                len(self._input_chemical_symbols), len(self._input_structure))
            raise ValueError(msg)
        else:
            chemical_symbols = copy.deepcopy(self._input_chemical_symbols)

        for i, symbols in enumerate(chemical_symbols):
            if len(symbols) != len(set(symbols)):
                raise ValueError(
                    'Found duplicates of allowed chemical symbols on site {}.'
                    ' allowed species on  site {}= {}'.format(i, i, symbols))

        if len([tuple(sorted(s)) for s in chemical_symbols if len(s) > 1]) == 0:
            raise ValueError('No active sites found')

        return chemical_symbols

    def _get_chemical_symbol_representation(self):
        """Returns a str version of the chemical symbols that is
        easier on the eyes.
        """
        sublattices = self.get_sublattices(self.primitive_structure)
        nice_str = []
        for sublattice in sublattices.active_sublattices:
            sublattice_symbol = sublattice.symbol

            nice_str.append('{} (sublattice {})'.format(
                list(sublattice.chemical_symbols), sublattice_symbol))
        return ', '.join(nice_str)

    def _get_string_representation(self,
                                   print_threshold: int = None,
                                   print_minimum: int = 10) -> str:
        """
        String representation of the cluster space that provides an overview of
        the orbits (order, radius, multiplicity etc) that constitute the space.

        Parameters
        ----------
        print_threshold
            if the number of orbits exceeds this number print dots
        print_minimum
            number of lines printed from the top and the bottom of the orbit
            list if `print_threshold` is exceeded

        Returns
        -------
        multi-line string
            string representation of the cluster space.
        """

        def repr_orbit(orbit, header=False):
            formats = {'order': '{:2}',
                       'radius': '{:8.4f}',
                       'multiplicity': '{:4}',
                       'index': '{:4}',
                       'orbit_index': '{:4}',
                       'multi_component_vector': '{:}',
                       'sublattices': '{:}'}
            s = []
            for name, value in orbit.items():
                str_repr = formats[name].format(value)
                n = max(len(name), len(str_repr))
                if header:
                    s += ['{s:^{n}}'.format(s=name, n=n)]
                else:
                    s += ['{s:^{n}}'.format(s=str_repr, n=n)]
            return ' | '.join(s)

        # basic information
        # (use largest orbit to obtain maximum line length)
        prototype_orbit = self.orbit_data[-1]
        width = len(repr_orbit(prototype_orbit))
        s = []  # type: List
        s += ['{s:=^{n}}'.format(s=' Cluster Space ', n=width)]
        s += [' chemical species: {}'
              .format(self._get_chemical_symbol_representation())]
        s += [' cutoffs: {}'.format(' '.join(['{:.4f}'.format(co)
                                              for co in self._cutoffs]))]
        s += [' total number of parameters: {}'.format(len(self))]
        t = ['{}= {}'.format(k, c)
             for k, c in self.get_number_of_orbits_by_order().items()]
        s += [' number of parameters by order: {}'.format('  '.join(t))]

        # table header
        s += [''.center(width, '-')]
        s += [repr_orbit(prototype_orbit, header=True)]
        s += [''.center(width, '-')]

        # table body
        index = 0
        orbit_list_info = self.orbit_data
        while index < len(orbit_list_info):
            if (print_threshold is not None and
                    len(self) > print_threshold and
                    index >= print_minimum and
                    index <= len(self) - print_minimum):
                index = len(self) - print_minimum
                s += [' ...']
            s += [repr_orbit(orbit_list_info[index])]
            index += 1
        s += [''.center(width, '=')]

        return '\n'.join(s)

    def __repr__(self) -> str:
        """ String representation. """
        return self._get_string_representation(print_threshold=50)

    def print_overview(self,
                       print_threshold: int = None,
                       print_minimum: int = 10) -> None:
        """
        Print an overview of the cluster space in terms of the orbits (order,
        radius, multiplicity etc).

        Parameters
        ----------
        print_threshold
            if the number of orbits exceeds this number print dots
        print_minimum
            number of lines printed from the top and the bottom of the orbit
            list if `print_threshold` is exceeded
        """
        print(self._get_string_representation(print_threshold=print_threshold,
                                              print_minimum=print_minimum))

    @property
    def orbit_data(self) -> List[dict]:
        """
        list of orbits with information regarding
        order, radius, multiplicity etc
        """
        data = []
        zerolet = OrderedDict([('index', 0),
                               ('order', 0),
                               ('radius', 0),
                               ('multiplicity', 1),
                               ('orbit_index', -1),
                               ('multi_component_vector', '.'),
                               ('sublattices', '.')])
        sublattices = self.get_sublattices(self.primitive_structure)
        data.append(zerolet)
        index = 1
        while index < len(self):
            cluster_space_info = self.get_cluster_space_info(index)
            orbit_index = cluster_space_info[0]
            mc_vector = cluster_space_info[1]
            orbit = self.get_orbit(orbit_index)
            rep_sites = orbit.get_representative_sites()
            orbit_sublattices = '-'.join(
                [sublattices[sublattices.get_sublattice_index(ls.index)].symbol
                 for ls in rep_sites])
            local_Mi = self.get_number_of_allowed_species_by_site(
                self._get_primitive_structure(), orbit.representative_sites)
            mc_vectors = orbit.get_mc_vectors(local_Mi)
            mc_permutations = self.get_multi_component_vector_permutations(
                mc_vectors, orbit_index)
            mc_index = mc_vectors.index(mc_vector)
            mc_permutations_multiplicity = len(mc_permutations[mc_index])
            cluster = self.get_orbit(orbit_index).get_representative_cluster()

            multiplicity = len(self.get_orbit(
                orbit_index).get_equivalent_sites())
            record = OrderedDict([('index', index),
                                  ('order', cluster.order),
                                  ('radius', cluster.radius),
                                  ('multiplicity', multiplicity *
                                   mc_permutations_multiplicity),
                                  ('orbit_index', orbit_index)])
            record['multi_component_vector'] = mc_vector
            record['sublattices'] = orbit_sublattices
            data.append(record)
            index += 1
        return data

    def get_number_of_orbits_by_order(self) -> OrderedDict:
        """
        Returns the number of orbits by order.

        Returns
        -------
        an ordered dictionary where keys and values represent order and number
        of orbits, respectively
        """
        count_orbits = {}  # type: dict[int, int]
        for orbit in self.orbit_data:
            k = orbit['order']
            count_orbits[k] = count_orbits.get(k, 0) + 1
        return OrderedDict(sorted(count_orbits.items()))

    def get_cluster_vector(self, structure: Atoms) -> np.ndarray:
        """
        Returns the cluster vector for a structure.

        Parameters
        ----------
        structure
            atomic configuration

        Returns
        -------
        the cluster vector
        """
        if not isinstance(structure, Atoms):
            raise TypeError('Input structure must be an ASE Atoms object')

        try:
            cv = _ClusterSpace.get_cluster_vector(self, Structure.from_atoms(structure))
        except Exception as e:
            self.assert_structure_compatibility(structure)
            raise(e)
        return cv

    def _prune_orbit_list(self, indices: List[int]) -> None:
        """
        Prunes the internal orbit list

        Parameters
        ----------
        indices
            indices to all orbits to be removed
        """
        size_before = len(self._orbit_list)

        self._prune_orbit_list_cpp(indices)
        for index in sorted(indices, reverse=True):
            self._orbit_list.remove_orbit(index)
        self._precompute_multi_component_vectors()

        size_after = len(self._orbit_list)
        assert size_before - len(indices) == size_after
        self._pruning_history.append(indices)

    @property
    def primitive_structure(self) -> Atoms:
        """ Primitive structure on which cluster space is based """
        structure = self._get_primitive_structure().to_atoms()
        # Decorate with the "real" symbols (instead of H, He, Li etc)
        for atom, symbols in zip(structure, self._primitive_chemical_symbols):
            atom.symbol = min(symbols)
        return structure

    @property
    def chemical_symbols(self) -> List[List[str]]:
        """ Species identified by their chemical symbols """
        return self._primitive_chemical_symbols.copy()

    @property
    def cutoffs(self) -> List[float]:
        """
        Cutoffs for different n-body clusters. The cutoff radius (in
        Angstroms) defines the largest interatomic distance in a
        cluster.
        """
        return self._cutoffs

    @property
    def orbit_list(self):
        """Orbit list that defines the cluster in the cluster space"""
        return self._orbit_list

    def get_possible_orbit_occupations(self, orbit_index: int) \
            -> List[List[str]]:
        """Returns possible occupation of the orbit.

        Parameters
        ----------
        orbit_index
        """
        orbit = self.orbit_list.orbits[orbit_index]

        indices = [
            lattice_site.index for lattice_site in orbit.representative_sites]

        allowed_species = [self.chemical_symbols[index] for index in indices]

        return list(itertools.product(*allowed_species))

    def get_sublattices(self, structure: Atoms) -> Sublattices:
        """
        Returns the sublattices of the input structure.

        Parameters
        ----------
        structure
            structure the sublattices are based on
        """
        sl = Sublattices(self.chemical_symbols, self.primitive_structure, structure)
        return sl

    def assert_structure_compatibility(self, structure: Atoms, vol_tol: float = 1e-5) -> None:
        """ Raises error if structure is not compatible with ClusterSpace.

        Todo
        ----
        Add check for if structure is relaxed.

        Parameters
        ----------
        structure
            structure to check if compatible with ClusterSpace
        """
        # check volume
        prim = self.primitive_structure
        vol1 = prim.get_volume() / len(prim)
        vol2 = structure.get_volume() / len(structure)
        if abs(vol1 - vol2) > vol_tol:
            raise ValueError('Volume per atom of structure does not match the volume of '
                             'ClusterSpace.primitive_structure')

        # check occupations
        sublattices = self.get_sublattices(structure)
        sublattices.assert_occupation_is_allowed(structure.get_chemical_symbols())

        # check pbc
        if not all(structure.pbc):
            raise ValueError('Input structure must have periodic boundary conditions')

    def is_supercell_self_correlated(self, structure: Atoms) -> bool:
        """
        Checks whether an structure has self-interactions via periodic
        boundary conditions.

        Parameters
        ----------
        structure
            structure to be tested

        Returns
        -------
        bool
            If True, the structure contains self-interactions via periodic
            boundary conditions, otherwise False.
        """
        ol = self.orbit_list.get_supercell_orbit_list(structure)
        orbit_indices = set()
        for orbit in ol.orbits:
            for sites in orbit.get_equivalent_sites():
                indices = tuple(sorted([site.index for site in sites]))
                if indices in orbit_indices:
                    return True
                else:
                    orbit_indices.add(indices)
        return False

    def write(self, filename: str) -> None:
        """
        Saves cluster space to a file.

        Parameters
        ---------
        filename
            name of file to which to write
        """

        with tarfile.open(name=filename, mode='w') as tar_file:

            # write items
            items = dict(cutoffs=self._cutoffs, chemical_symbols=self._input_chemical_symbols,
                         pruning_history=self._pruning_history)
            temp_file = tempfile.TemporaryFile()
            pickle.dump(items, temp_file)
            temp_file.seek(0)
            tar_info = tar_file.gettarinfo(arcname='items', fileobj=temp_file)
            tar_file.addfile(tar_info, temp_file)
            temp_file.close()

            # write structure
            temp_file = tempfile.NamedTemporaryFile()
            ase_write(temp_file.name, self._input_structure, format='json')
            temp_file.seek(0)
            tar_info = tar_file.gettarinfo(arcname='atoms', fileobj=temp_file)
            tar_file.addfile(tar_info, temp_file)

    @staticmethod
    def read(filename: str):
        """
        Reads cluster space from filename.

        Parameters
        ---------
        filename
            name of file from which to read cluster space
        """
        if isinstance(filename, str):
            tar_file = tarfile.open(mode='r', name=filename)
        else:
            tar_file = tarfile.open(mode='r', fileobj=filename)

        # read items
        items = pickle.load(tar_file.extractfile('items'))

        # read structure
        temp_file = tempfile.NamedTemporaryFile()
        temp_file.write(tar_file.extractfile('atoms').read())
        temp_file.seek(0)
        structure = ase_read(temp_file.name, format='json')

        tar_file.close()
        cs = ClusterSpace(structure=structure, cutoffs=items['cutoffs'],
                          chemical_symbols=items['chemical_symbols'])
        for indices in items['pruning_history']:
            cs._prune_orbit_list(indices)
        return cs

    def copy(self):
        """ Returns copy of ClusterSpace instance. """
        structure = self._input_structure
        cutoffs = self._cutoffs
        chemical_symbols = self._input_chemical_symbols
        cs_copy = ClusterSpace(structure, cutoffs, chemical_symbols)
        for indices in self._pruning_history:
            cs_copy._prune_orbit_list(indices)
        return cs_copy


def get_singlet_info(structure: Atoms,
                     return_cluster_space: bool = False):
    """
    Retrieves information concerning the singlets in the input structure.

    Parameters
    ----------
    structure
        atomic configuration
    return_cluster_space
        if True return the cluster space created during the process

    Returns
    -------
    list of dicts
        each dictionary in the list represents one orbit
    ClusterSpace object (optional)
        cluster space created during the process
    """
    if not isinstance(structure, Atoms):
        raise TypeError('Input configuration must be an ASE Atoms object'
                        ', not type {}'.format(type(structure)))

    # create dummy species and cutoffs
    chemical_symbols = ['H', 'He']
    cutoffs = [0.0]

    cs = ClusterSpace(structure, cutoffs, chemical_symbols)

    singlet_data = []

    for i in range(1, len(cs)):
        cluster_space_info = cs.get_cluster_space_info(i)
        orbit_index = cluster_space_info[0]
        cluster = cs.get_orbit(orbit_index).get_representative_cluster()
        multiplicity = len(cs.get_orbit(orbit_index).get_equivalent_sites())
        assert len(cluster) == 1, \
            'Cluster space contains higher-order terms (beyond singlets)'

        singlet = {}
        singlet['orbit_index'] = orbit_index
        singlet['sites'] = cs.get_orbit(orbit_index).get_equivalent_sites()
        singlet['multiplicity'] = multiplicity
        singlet['representative_site'] = cs.get_orbit(
            orbit_index).get_representative_sites()
        singlet_data.append(singlet)

    if return_cluster_space:
        return singlet_data, cs
    else:
        return singlet_data


def get_singlet_configuration(structure: Atoms,
                              to_primitive: bool = False) -> Atoms:
    """
    Returns the atomic configuration occupied with a different species for
    each Wyckoff site. This is useful for visualization and analysis.

    Parameters
    ----------
    structure
        atomic configuration
    to_primitive
        if True the input structure will be reduced to its primitive unit cell
        before processing

    Returns
    -------
    structure with singlets highlighted by different chemical species
    """
    from ase.data import chemical_symbols
    assert isinstance(structure, Atoms), \
        'input configuration must be an ASE Atoms object'
    cluster_data, cluster_space = get_singlet_info(structure,
                                                   return_cluster_space=True)

    if to_primitive:
        structure_singlet = cluster_space.primitive_structure
        for singlet in cluster_data:
            for site in singlet['sites']:
                symbol = chemical_symbols[singlet['orbit_index'] + 1]
                atom_index = site[0].index
                structure_singlet[atom_index].symbol = symbol
    else:
        structure_singlet = structure.copy()
        orbit_list_supercell = \
            cluster_space._orbit_list.get_supercell_orbit_list(structure_singlet)
        for singlet in cluster_data:
            for site in singlet['sites']:
                symbol = chemical_symbols[singlet['orbit_index'] + 1]
                sites = orbit_list_supercell.get_orbit(
                    singlet['orbit_index']).get_equivalent_sites()
                for lattice_site in sites:
                    k = lattice_site[0].index
                    structure_singlet[k].symbol = symbol

    return structure_singlet


def view_singlets(structure: Atoms, to_primitive: bool = False):
    """
    Visualizes singlets in a structure using the ASE graphical user interface.

    Parameters
    ----------
    structure
        atomic configuration
    to_primitive
        if True the input structure will be reduced to its primitive unit cell
        before processing
    """
    from ase.visualize import view
    assert isinstance(structure, Atoms), \
        'input configuration must be an ASE Atoms object'
    structure_singlet = get_singlet_configuration(structure, to_primitive=to_primitive)
    view(structure_singlet)



import itertools
import random
from typing import List
import numpy as np
from mchammer.ensembles import TargetClusterVectorAnnealing
from mchammer.calculators import (TargetVectorCalculator,
                                  compare_cluster_vectors)
from icet.tools import (enumerate_structures,
                        enumerate_supercells)
from icet import ClusterSpace
from ase import Atoms
from ase.data import chemical_symbols as periodic_table


def generate_target_structure(cluster_space: ClusterSpace, max_size: int,
                              target_concentrations: dict,
                              target_cluster_vector: List[float],
                              include_smaller_cells: bool = True,
                              T_start: float = 5.0, T_stop: float = 0.001,
                              n_steps: float = None,
                              optimality_weight: float = 1.0,
                              random_seed: int = None,
                              tol: float = 1e-5) -> Atoms:
    """
    Given a ``cluster_space`` and a ``target_cluster_vector``, generate
    a structure that as closely as possible matches that cluster vector.
    The search is performed among all inequivalent supercells shapes up
    to a certain size.

    Internally the function uses a simulated annealing algorithm and the
    difference between two cluster vectors is calculated with the
    measure suggested by A. van de Walle et al. in Calphad **42**, 13-18
    (2013) [WalTiwJon13]_ (for more information, see
    :class:`mchammer.calculators.TargetVectorCalculator`).

    Parameters
    ----------
    cluster_space
        a cluster space defining the lattice to be occupied
    max_size
        maximum supercell size
    target_concentrations
        concentration of each species in the target structure (for
        example ``{'Ag': 0.5, 'Pd': 0.5}``

        Concentrations are always expressed with respect to all atoms in
        the supercell, which implies that the sum of all concentrations
        should always be 1. In the case of multiple sublattices, a valid
        specification would thus be ``{'Au': 0.25, 'Pd': 0.25, 'H': 0.1,
        'V': 0.4}``.
    target_cluster_vector
        cluster vector that the generated structure should match as
        closely as possible
    include_smaller_cells
        if True, search among all supercell sizes including
        ``max_size``, else search only among those exactly matching
        ``max_size``
    T_start
        artificial temperature at which the simulated annealing starts
    T_stop
        artifical temperature at which the simulated annealing stops
    n_steps
        total number of Monte Carlo steps in the simulation
    optimality_weight
        controls weighting :math:`L` of perfect correlations, see
        :class:`mchammer.calculators.TargetVectorCalculator`
    random_seed
        seed for the random number generator used in the
        Monte Carlo simulation
    tol
        Numerical tolerance
    """
    _validate_concentrations(target_concentrations, cluster_space)

    supercells = []
    calculators = []
    if include_smaller_cells:
        sizes = range(1, max_size + 1)
    else:
        sizes = [max_size]

    for size in sizes:
        # For efficiency, make a first check that the current size is
        # commensurate with all concentrations (example: {'Au': 0.5,
        # 'Pd': 0.5} would not be commensurate with a supercell with 3
        # atoms). This step does *not* ensure that the concentrations
        # are commensurate with all sublattices (that is checked
        # internally later on).
        natoms = size * len(cluster_space.primitive_structure)
        if not _concentrations_fit_atom_count(natoms, target_concentrations):
            continue

        # Loop over all inequivalent supercells and intialize
        # them with a "random" occupation of symbols that
        # fulfill the target concentrations
        for supercell in enumerate_supercells(cluster_space.primitive_structure,
                                              [size]):
            occupy_structure_randomly(supercell, cluster_space,
                                      target_concentrations)
            supercells.append(supercell)
            calculators.append(TargetVectorCalculator(supercell, cluster_space,
                                                      target_cluster_vector,
                                                      optimality_weight=optimality_weight,
                                                      optimality_tol=tol))

    ens = TargetClusterVectorAnnealing(structure=supercells, calculators=calculators,
                                       T_start=T_start, T_stop=T_stop,
                                       random_seed=random_seed)
    return ens.generate_structure(number_of_trial_steps=n_steps)


def generate_sqs(cluster_space: ClusterSpace, max_size: int,
                 target_concentrations: dict,
                 include_smaller_cells: bool = True,
                 T_start: float = 5.0, T_stop: float = 0.001,
                 n_steps: float = None,
                 optimality_weight: float = 1.0,
                 random_seed: int = None,
                 tol: float = 1e-5) -> Atoms:
    """
    Given a ``cluster_space``, generate a special quasirandom structure
    (SQS), i.e., a structure that for a given supercell size provides
    the best possible approximation to a random alloy [ZunWeiFer90]_.

    In the present case, this means that the generated structure will
    have a cluster vector that as closely as possible matches the
    cluster vector of an infintely large randomly occupated supercell.
    Internally the function uses a simulated annealing algorithm and the
    difference between two cluster vectors is calculated with the
    measure suggested by A. van de Walle et al. in Calphad **42**, 13-18
    (2013) [WalTiwJon13]_ (for more information, see
    :class:`mchammer.calculators.TargetVectorCalculator`).

    Parameters
    ----------
    cluster_space
        a cluster space defining the lattice to be occupated
    max_size
        maximum supercell size
    target_concentrations
        concentration of each species in the target structure (for
        example ``{'Ag': 0.5, 'Pd': 0.5}``

        Concentrations are always expressed with respect to all atoms in
        the supercell, which implies that the sum of all concentrations
        should always be 1. In the case of multiple sublattices, a valid
        specification would thus be ``{'Au': 0.25, 'Pd': 0.25, 'H': 0.1,
        'V': 0.4}``.
    include_smaller_cells
        if True, search among all supercell sizes including
        ``max_size``, else search only among those exactly matching
        ``max_size``
    T_start
        artificial temperature at which the simulated annealing starts
    T_stop
        artifical temperature at which the simulated annealing stops
    n_steps
        total number of Monte Carlo steps in the simulation
    optimality_weight
        controls weighting :math:`L` of perfect correlations, see
        :class:`mchammer.calculators.TargetVectorCalculator`
    random_seed
        seed for the random number generator used in the
        Monte Carlo simulation
    tol
        Numerical tolerance
    """

    sqs_vector = _get_sqs_cluster_vector(cluster_space=cluster_space,
                                         target_concentrations=target_concentrations)
    return generate_target_structure(cluster_space=cluster_space,
                                     max_size=max_size,
                                     target_concentrations=target_concentrations,
                                     target_cluster_vector=sqs_vector,
                                     include_smaller_cells=include_smaller_cells,
                                     T_start=T_start, T_stop=T_stop,
                                     n_steps=n_steps,
                                     optimality_weight=optimality_weight,
                                     random_seed=random_seed,
                                     tol=tol)


def generate_sqs_by_enumeration(cluster_space: ClusterSpace, max_size: int,
                                target_concentrations: dict,
                                include_smaller_cells: bool = True,
                                optimality_weight: float = 1.0,
                                tol: float = 1e-5) -> Atoms:
    """
    Given a ``cluster_space``, generate a special quasirandom structure
    (SQS), i.e., a structure that for a given supercell size provides
    the best possible approximation to a random alloy [ZunWeiFer90]_.

    In the present case, this means that the generated structure will
    have a cluster vector that as closely as possible matches the
    cluster vector of an infintely large randomly occupied supercell.
    Internally the function uses a simulated annealing algorithm and the
    difference between two cluster vectors is calculated with the
    measure suggested by A. van de Walle et al. in Calphad **42**, 13-18
    (2013) [WalTiwJon13]_ (for more information, see
    :class:`mchammer.calculators.TargetVectorCalculator`).

    This functions generates SQS cells by exhaustive enumeration, which
    means that the generated SQS cell is guaranteed to be optimal with
    regard to the specified measure and cell size.

    Parameters
    ----------
    cluster_space
        a cluster space defining the lattice to be occupied
    max_size
        maximum supercell size
    target_concentrations
        concentration of each species in the target structure (for
        example ``{'Ag': 0.5, 'Pd': 0.5}``

        Concentrations are always expressed with respect to all atoms in
        the supercell, which implies that the sum of all concentrations
        should always be 1. In the case of multiple sublattices, a valid
        specification would thus be ``{'Au': 0.25, 'Pd': 0.25, 'H': 0.1,
        'V': 0.4}``.
    include_smaller_cells
        if True, search among all supercell sizes including
        ``max_size``, else search only among those exactly matching
        ``max_size``
    optimality_weight
        controls weighting :math:`L` of perfect correlations, see
        :class:`mchammer.calculators.TargetVectorCalculator`
    tol
        Numerical tolerance
    """
    _validate_concentrations(target_concentrations, cluster_space)
    sqs_vector = _get_sqs_cluster_vector(cluster_space=cluster_space,
                                         target_concentrations=target_concentrations)
    # Translate concentrations to the format required for concentration
    # restricted enumeration
    cr = {key: (value, value) for key, value in target_concentrations.items()}

    orbit_data = cluster_space.orbit_data
    best_score = 1e9

    if include_smaller_cells:
        sizes = range(1, max_size + 1)
    else:
        sizes = [max_size]

    # Enumerate and calculate score for each structuer
    for structure in enumerate_structures(cluster_space.primitive_structure,
                                          sizes,
                                          cluster_space.chemical_symbols,
                                          concentration_restrictions=cr):
        cv = cluster_space.get_cluster_vector(structure)
        score = compare_cluster_vectors(cv_1=cv, cv_2=sqs_vector,
                                        orbit_data=orbit_data,
                                        optimality_weight=optimality_weight,
                                        tol=tol)

        if score < best_score:
            best_score = score
            best_structure = structure
    return best_structure


def occupy_structure_randomly(structure: Atoms, cluster_space: ClusterSpace,
                              target_concentrations: dict):
    """
    Occupy a structure with quasirandom order but fulfilling
    ``target_concentrations``.

    Parameters
    ----------
    structure
        ASE Atoms object that will be occupied randomly
    cluster_space
        cluster space (needed as it carries information about sublattices)
    target_concentrations
        concentration of each species in the target structure (for
        example ``{'Ag': 0.5, 'Pd': 0.5}``

        Concentrations are always expressed with respect to all atoms in
        the supercell, which implies that the sum of all concentrations
        should always be 1. In the case of multiple sublattices, a valid
        specification would thus be ``{'Au': 0.25, 'Pd': 0.25, 'H': 0.1,
        'V': 0.4}``.
    """
    if not _concentrations_fit_structure(structure, cluster_space, target_concentrations):
        raise ValueError('Structure with {} atoms cannot accomodate '
                         'target concentrations {}'.format(len(structure),
                                                           target_concentrations))

    # symbols_all will hold chemical_symbols of all sublattices
    symbols_all = [''] * len(structure)
    for sublattice in cluster_space.get_sublattices(structure):
        symbols = []  # chemical_symbols in one sublattice
        for chemical_symbol in sublattice.chemical_symbols:
            n_symbol = int(round(len(structure) * target_concentrations[chemical_symbol]))
            symbols += [chemical_symbol] * n_symbol

        # Should not happen but you never know
        assert len(symbols) == len(sublattice.indices)

        # Shuffle to introduce randomness
        random.shuffle(symbols)

        # Assign symbols to the right indices
        for symbol, lattice_site in zip(symbols, sublattice.indices):
            symbols_all[lattice_site] = symbol

    assert symbols_all.count('') == 0
    structure.set_chemical_symbols(symbols_all)


def _validate_concentrations(concentrations: dict,
                             cluster_space: ClusterSpace,
                             tol: float = 1e-5):
    """
    Validates concentration specification against a cluster space
    (raises `ValueError` if they do not match).

    Parameters
    ----------
    concentrations
        concentration specification
    cluster_space
        cluster space to check against
    tol
        Numerical tolerance
    """
    # Concentrations should sum to 1
    conc_sum = sum(list(concentrations.values()))
    if abs(conc_sum - 1.0) > tol:
        raise ValueError('Concentrations must sum up '
                         'to 1 (not {})'.format(conc_sum))

    # Symbols need to match
    # Flatten chemical symbols from cluster space
    chemical_symbols = set(sym for syms in cluster_space.chemical_symbols for sym in syms)
    if chemical_symbols != set(concentrations.keys()):
        raise ValueError('Chemical symbols in cluster space ({}) are '
                         'not the same as those in the specified '
                         'concentrations ({})'.format(set(chemical_symbols),
                                                      set(concentrations.keys())))

    # Concentrations for each sublattice need to sum up to
    # its proportion of the total structure
    for sublattice in cluster_space.get_sublattices(cluster_space.primitive_structure):
        part = len(sublattice.indices) / len(cluster_space.primitive_structure)
        conc = sum([concentrations[sym]
                    for sym in sublattice.chemical_symbols])
        if abs(conc - part) > tol:
            raise ValueError('Concentrations must be specified such that '
                             'concentrations per element and per sublattice '
                             'match, sublattice with species {} contains a '
                             'fraction {:.2f} of all atoms, but the sum of '
                             'the concentrations of those symbols is '
                             '{:.2f}.'.format(sublattice.chemical_symbols,
                                              part,
                                              conc))


def _concentrations_fit_atom_count(atom_count: int,
                                   concentrations: dict,
                                   tol: float = 1e-5) -> bool:
    """
    Check if specified concentrations are commensurate with a
    certain number of atoms.

    Parameters
    ----------
    atom_count
        How many atoms?
    concentrations
        What concentrations? Example: ``{'Ag': 0.3, 'Au': 0.7}``
    tol
        Numerical tolerance
    """
    for conc in concentrations.values():
        n_symbol = conc * atom_count
        if abs(int(round(n_symbol)) - n_symbol) > tol:
            return False
    return True


def _concentrations_fit_structure(structure: Atoms,
                                  cluster_space: ClusterSpace,
                                  concentrations: dict,
                                  tol: float = 1e-5) -> bool:
    """
    Check if specified concentrations are commensurate with a
    certain supercell (including sublattices)

    Parameters
    ----------
    structure
        ASE Atoms object to be checked
    cluster_space
        corresponding cluster space
    concentrations
        What concentrations? Example: ``{'Ag': 0.3, 'Au': 0.7}``
    tol
        Numerical tolerance
    """
    # Check that concentrations are OK in each sublattice
    for sublattice in cluster_space.get_sublattices(structure):
        # Since concentrations are specified relative to all atoms,
        # we must rescale when we check against the size of a sublattice
        scaling = len(structure) / len(sublattice.indices)
        concentrations_sub = {elem: scaling * concentrations[elem]
                              for elem in sublattice.chemical_symbols}
        if not _concentrations_fit_atom_count(len(sublattice.indices),
                                              concentrations_sub,
                                              tol=tol):
            return False
    return True


def _get_sqs_cluster_vector(cluster_space: ClusterSpace,
                            target_concentrations: dict) -> np.ndarray:
    """
    Get the SQS vector for a certain cluster space and certain
    concentration. Here SQS vector refers to the cluster vector of an
    infintely large supercell with random occupation.

    Parameters
    ----------
    cluster_space
        The kind of lattice to be occupied
    target_concentrations
        concentration of each species in the target structure (for
        example `{'Ag': 0.5, 'Pd': 0.5}`

        Concentrations are always expressed with respect to all atoms in
        the supercell, which implies that the sum of all concentrations
        should always be 1. In the case of multiple sublattices, a valid
        specification would thus be `{'Au': 0.25, 'Pd': 0.25, 'H': 0.1,
        'V': 0.4}`.
    """
    _validate_concentrations(target_concentrations, cluster_space)

    sublattice_to_index = {letter: index for index,
                           letter in enumerate('ABCDEFGHIJKLMNOPQRSTVUWWXYZ')}
    all_sublattices = cluster_space.get_sublattices(
        cluster_space.primitive_structure)

    # Make a map from chemical symbol to integer, later on used
    # for evaluating cluster functions.
    # Internally, icet sorts species according to atomic numbers.
    # Also check that each symbol only occurs in one sublattice.
    symbol_to_integer_map = {}
    found_species = []
    for sublattice in all_sublattices:
        syms = sublattice.chemical_symbols
        atomic_numbers = [periodic_table.index(sym) for sym in syms]
        for i, species in enumerate(sorted(atomic_numbers)):
            if species in found_species:
                raise ValueError('Each chemical symbol can only occur on one '
                                 'sublattice, {} occurred more than '
                                 'once.'.format(periodic_table[species]))
            found_species.append(species)
            symbol_to_integer_map[periodic_table[species]] = i

    # Target concentrations refer to all atoms, but probabilities only
    # to the sublattice.
    probabilities = {}
    for sublattice in all_sublattices:
        mult_factor = len(cluster_space.primitive_structure) / \
            len(sublattice.indices)
        for symbol in sublattice.chemical_symbols:
            probabilities[symbol] = target_concentrations[symbol] * mult_factor

    # For every orbit, calculate average cluster function
    cv = [1.0]
    for orbit in cluster_space.orbit_data:
        if orbit['order'] < 1:
            continue

        # What sublattices are there in this orbit?
        sublattices = [all_sublattices[sublattice_to_index[letter]]
                       for letter in orbit['sublattices'].split('-')]

        # What chemical symbols do these sublattices refer to?
        symbol_groups = [
            sublattice.chemical_symbols for sublattice in sublattices]

        # How many allowed species in each of those sublattices?
        nbr_of_allowed_species = [len(symbol_group)
                                  for symbol_group in symbol_groups]

        # Calculate contribtion from every possible combination of
        # symbols weighted with their probability
        cluster_product_average = 0
        for symbols in itertools.product(*symbol_groups):
            cluster_product = 1
            for i, symbol in enumerate(symbols):
                mc_vector_component = orbit['multi_component_vector'][i]
                species_i = symbol_to_integer_map[symbol]
                prod = cluster_space.evaluate_cluster_function(nbr_of_allowed_species[i],
                                                               mc_vector_component,
                                                               species_i)
                cluster_product *= probabilities[symbol] * prod
            cluster_product_average += cluster_product
        cv.append(cluster_product_average)
    return np.array(cv)



"""
This module provides a Python interface to the PermutationMatrix
class with supplementary functions.
"""

from typing import List, Tuple

import numpy as np
import spglib

from ase import Atoms
from _icet import PermutationMatrix
from icet.core.lattice_site import LatticeSite
from icet.core.neighbor_list import NeighborList
from icet.core.structure import Structure
from icet.io.logging import logger
from icet.tools.geometry import (ase_atoms_to_spglib_cell,
                                 get_fractional_positions_from_neighbor_list,
                                 get_primitive_structure)

logger = logger.getChild('permutation_matrix')


def permutation_matrix_from_structure(structure: Atoms, cutoff: float,
                                      find_prim: bool = True) \
        -> Tuple[np.ndarray, Structure, NeighborList]:
    """Sets up a list of permutation maps from an Atoms object.

    Parameters
    ----------
    structure
        input structure
    cutoff
        cutoff radius
    find_primitive
        if True the symmetries of the primitive structure will be employed

    Returns
    -------
    the tuple that is returned comprises the permutation matrix, the
    primitive structure, and the neighbor list
    """

    structure = structure.copy()

    structure_prim = structure
    if find_prim:
        structure_prim = get_primitive_structure(structure)

    logger.debug('Size of primitive structure: {}'.format(len(structure_prim)))

    structure_as_tuple = ase_atoms_to_spglib_cell(structure_prim)

    # get symmetry information
    symmetry = spglib.get_symmetry(structure_as_tuple)
    translations = symmetry['translations']
    rotations = symmetry['rotations']

    # set up a permutation map object
    permutation_matrix = PermutationMatrix(translations, rotations)

    # create neighbor_lists from the different cutoffs
    prim_icet_structure = Structure.from_atoms(structure_prim)
    neighbor_list = NeighborList(cutoff)
    neighbor_list.build(prim_icet_structure)

    # get fractional positions for neighbor_list
    frac_positions = get_fractional_positions_from_neighbor_list(
        prim_icet_structure, neighbor_list)

    logger.debug('Number of fractional positions:'
                 ' {}'.format(len(frac_positions)))
    if frac_positions is not None:
        permutation_matrix.build(frac_positions)

    return permutation_matrix, prim_icet_structure, neighbor_list


def _get_lattice_site_permutation_matrix(structure: Structure,
                                         permutation_matrix: PermutationMatrix,
                                         prune: bool = True) -> np.ndarray:
    """
    Returns a transformed permutation matrix with lattice sites as entries
    instead of fractional coordinates.

    Parameters
    ----------
    structure
        primitive atomic icet structure
    permutation_matrix
        permutation matrix with fractional coordinates format entries
    prune
        if True the permutation matrix will be pruned

    Returns
    -------
    Permutation matrix in a row major order with lattice site format entries
    """
    pm_frac = permutation_matrix.get_permuted_positions()

    pm_lattice_sites = []
    for row in pm_frac:
        positions = _fractional_to_cartesian(row, structure.cell)
        lat_neighbors = []
        if np.all(structure.pbc):
            lat_neighbors = \
                structure.find_lattice_sites_by_positions(positions)
        else:
            for pos in positions:
                try:
                    lat_neighbor = \
                        structure.find_lattice_site_by_position(pos)
                except RuntimeError:
                    continue
                lat_neighbors.append(lat_neighbor)
        if lat_neighbors is not None:
            pm_lattice_sites.append(lat_neighbors)
        else:
            logger.warning('Unable to transform any element in a column of the'
                           ' fractional permutation matrix to lattice site')
    if prune:
        logger.debug('Size of columns of the permutation matrix before'
                     ' pruning {}'.format(len(pm_lattice_sites)))

        pm_lattice_sites = _prune_permutation_matrix(pm_lattice_sites)

        logger.debug('Size of columns of the permutation matrix after'
                     ' pruning {}'.format(len(pm_lattice_sites)))

    return pm_lattice_sites


def _prune_permutation_matrix(permutation_matrix: List[List[LatticeSite]]):
    """
    Prunes the matrix so that the first column only contains unique elements.

    Parameters
    ----------
    permutation_matrix
        permutation matrix with LatticeSite type entries
    """

    for i in range(len(permutation_matrix)):
        for j in reversed(range(len(permutation_matrix))):
            if j <= i:
                continue
            if permutation_matrix[i][0] == permutation_matrix[j][0]:
                permutation_matrix.pop(j)
                logger.debug('Removing duplicate in permutation matrix'
                             'i: {} j: {}'.format(i, j))

    return permutation_matrix


def _fractional_to_cartesian(fractional_coordinates: List[List[float]],
                             cell: np.ndarray) -> List[float]:
    """
    Converts cell metrics from fractional to cartesian coordinates.

    Parameters
    ----------
    fractional_coordinates
        list of fractional coordinates

    cell
        cell metric
    """
    cartesian_coordinates = [np.dot(frac, cell)
                             for frac in fractional_coordinates]
    return cartesian_coordinates



"""
This module collects a number of different tools, e.g., for
structure generation and analysis.
"""

from .convex_hull import ConvexHull
from .structure_enumeration import (enumerate_structures,
                                    enumerate_supercells)
from .geometry import (get_wyckoff_sites,
                       get_primitive_structure)
from .structure_mapping import map_structure_to_reference

__all__ = ['ConvexHull',
           'enumerate_structures',
           'enumerate_supercells',
           'get_primitive_structure',
           'get_wyckoff_sites',
           'map_structure_to_reference']



import itertools
from collections import OrderedDict


class LabelingGenerator():
    """
    Object used to generate all possible permutations of species on a given
    set of sites. If concentration restrictions are not specified, the
    generation will be a simple itertools.product loop.

    If concentrations are specified, the approach is a bit more elaborate.
    Take as an example a system with four sites, where Pd and Au are allowed
    on two and H and V on the others. `iter_species` will then look something
    like this: `[(0, 1), (0, 1), (2, 3), (2, 3)]`. The algorithm proceeds as
    follows:

    (1) Identify all unique species of `iter_species`, hereafter called
    `site_groups`. In this example they will be `(0, 1)` and `(2, 3)`. These
    unqiue species are saved in `site_groups`.

    (2) For each unique `site_group` and a given cell size, construct all
    (combinations of the species in the site group. If the cell size is 2,
    i.e. a supercell twice as big as the primitive cell, the possible
    (combinations for the `(0, 1)`-`site_group` will be `(0, 0, 0, 0)`, `(0,
    0, 0, 1)`, `(0, 0, 1, 1)` and `(1, 1, 1, 1)`. Order does not matter and
    concentration restrictions are not respected at this point.

    (3) Construct all "products" of the combinations of the unique site
    groups. In our example, the first ones are `[(0, 0, 0, 0), (2, 2, 2,
    2)]`, `[(0, 0, 0, 1), (2, 2, 2, 2)]` etc. Products that do not respect the
    concentration restrictions are discarded at this point.

    (4) For each "product", construct the permutations for each constituent
    `site_group`. For `[(0, 0, 0, 1), (2, 2, 2, 2)]` these would be
    `[(0, 0, 0, 1), (2, 2, 2, 2)]`, `[(0, 0, 1, 0), (2, 2, 2, 2)]`,
    `[(0, 1, 0, 0), (2, 2, 2, 2)]` and `[(1, 0, 0, 0), (2, 2, 2, 2)]`.
    If the second group was not all 2, there would of course be many more
    such permutations.

    (5) Sort the result according to the expectations of the structure
    enumeration (if we call the site group (0, 1) "A" and site group
    (2, 3) "B", the labeling should be AABBAABBAABB etc, i.e., one cell
    at a time). In the first permutation of (4), the resulting labeling
    will thus be `(0, 0, 2, 2, 0, 1, 2, 2)`.

    Parameters
    ----------
    iter_species : list of tuples of ints
        allowed species on each site
    concentrations : dict
        keys represent species (integers), values specify concentration ranges
    tol : float
        tolerance parameter used when comparing concentrations

    Attribtues
    ----------
    site_groups : OrderedDict
        keys are unique iter_species, the values are SiteGroup objects
        corresponding to that iter_species
    """

    def __init__(self, iter_species, concentrations, tol=1e-5):
        self.iter_species = iter_species
        self.concentrations = concentrations
        self.tol = tol

        if self.concentrations:
            self.site_groups = OrderedDict()
            count = 0
            for iter_species_key in iter_species:
                if iter_species_key in self.site_groups.keys():
                    self.site_groups[iter_species_key].multiplicity += 1
                else:
                    self.site_groups[iter_species_key] = SiteGroup(
                        iter_species_key, count)
                    count += 1

    def yield_labelings(self, ncells):
        """
        Yield labelings that comply with the concentration restrictions and,
        with `ncells` primitive cells.

        Parameters
        ----------
        ncells : int
            Size of supercell

        Yields
        ------
        tuple of ints
            Labeling
        """

        if self.concentrations:
            for site_group in self.site_groups.values():
                site_group.compute_all_combinations(ncells)
            for product in self.yield_products(ncells):
                for labeling in self.yield_permutations(product, 0):
                    yield self.sort_labeling(labeling, ncells)
        else:
            for labeling in itertools.product(*self.iter_species * ncells):
                yield labeling

    def yield_products(self, ncells):
        """
        Yield combinations (or rather products in the itertools terminology)
        of decorated site group combinations that comply with the concentration
        restrictions.

        Returns
        -------
        tuple of tuples of ints
            all unique combinations (products) of unordered iter_species
            combinations,
        """
        natoms = len(self.iter_species) * ncells
        combinations = [sg.combinations for sg in self.site_groups.values()]
        for product in itertools.product(*combinations):
            counts = {species: 0 for species in self.concentrations.keys()}
            for species_group in product:
                for species in self.concentrations.keys():
                    counts[species] += species_group.count(species)
            conc_restr_violation = False
            for species, conc_range in self.concentrations.items():
                if counts[species] / natoms < conc_range[0] - self.tol or \
                   counts[species] / natoms > conc_range[1] + self.tol:
                    conc_restr_violation = True
                    break
            if not conc_restr_violation:
                yield product

    def yield_unique_permutations(self, unique_species, permutation,
                                  position):
        """
        Recursively generate all _unique_ permutations of a set of species
        with given multiplicities. The algorithm is inspired by the one given
        at https://stackoverflow.com/a/6285203/6729551

        Parameters
        ----------
        unique_species : dict
            keys represent species,values the corresponding multiplicity
        permutation : list of ints
            permutation in process; should have length equal to the sum of all
            multiplicities
        position : int
            position currently processed; should be the index of the last
            species of `permutation` upon initialization

        Yields
        ------
        tuple with ints
            permutation where each species occurs according to the multiplicity
            specified in `unique_species`
        """
        if position < 0:
            # Finish recursion
            yield tuple(permutation)
        else:
            for species, occurrences in unique_species.items():
                # Add if the multiplicity allows
                if occurrences > 0:
                    permutation[position] = species
                    unique_species[species] -= 1
                    for perm in self.yield_unique_permutations(unique_species,
                                                               permutation,
                                                               position - 1):
                        yield perm
                    unique_species[species] += 1

    def yield_permutations(self, product, position):
        """
        Recursively generate all combinations of unique permutations of the
        tuples in `product`.

        Parameters
        ----------
        product : tuple of tuples of ints
            species allowed for each site group
        position : int
            keeps track of the position where recursion occurs; set to 0 upon
            initialization.

        Yields
        ------
        list of tuples of ints
            unique combinations of unique permutations, ordered by site group
        """
        unique_species = {species: product[position].count(species)
                          for species in set(product[position])}
        natoms = len(product[position])
        for permutation in self.yield_unique_permutations(unique_species,
                                                          [0] * natoms,
                                                          natoms - 1):
            if position == len(product) - 1:
                yield [permutation]
            else:
                for permutation_rest in self.yield_permutations(product,
                                                                position + 1):
                    yield [permutation] + permutation_rest

    def sort_labeling(self, labeling, ncells):
        """
        The elements in labeling are now given in site groups. We want just
        one labeling, ordered by site group, but one primitive cell at a time.
        So if iter_species given upon initialization was
        `[(0, 1), (2), (0, 1)]` and `ncells=2`, the current labeling has two
        tuples, the first corresponding to the `(0, 1)` site group, with 8
        elements in total, and the second corresponding to the `(2)` site group
        and with 2 elements in total. Now, we reorder it so that the elements
        are ordered according to `[(0, 1), (2), (0, 1), (0, 1), (2), (0, 1)]`.

        Parameters
        ----------
        labeling : list of tuple of ints
            As given by yield_permutations
        ncells : int
            Size of supercell

        Returns
        -------
        tuple of ints
            Labeling properly sorted, ready to be given to the enumeration code
        """
        sorted_labeling = [0] * len(self.iter_species * ncells)
        count = 0
        for _ in range(ncells):
            for iter_species in self.iter_species:

                # Find the site group corresponding to the iter species
                site_group = self.site_groups[iter_species]

                # Find the right element by checking (1) where the
                # proper site_group is in the unsorted labeling and
                # (2) which element is next in turn
                sorted_labeling[count] = labeling[
                    site_group.position][site_group.next_to_add]
                count += 1
                site_group.next_to_add += 1

        # Reset site group counters
        for site_group in self.site_groups.values():
            assert site_group.next_to_add == len(labeling[site_group.position])
            site_group.next_to_add = 0
        return tuple(sorted_labeling)


class SiteGroup():
    """
    Keeps track of a group of sites that have the same allowed species.
    That is a site group could correspond to all sites on which species 0 and
    1 are allowed, and the number of such sites is stored in the `multiplicity`
    attribute..

    Parameters
    ----------
    iter_species : tuple of ints
        allowed species on these sites
    position : int
        helps to keep track of when the first group occurred; the first
        site group encountered will have position = 0, the next 1 etc.

    Attributes
    ----------
    multiplicity : int
        multiplicity if the group, i.e., how many sites that have this
        iter_species
    next_to_add : int
        helper attribute that keeps track of which site in this group is the
        next to be added when the elements of a labeling are to be sorted
    """

    def __init__(self, iter_species, position):
        self.iter_species = iter_species
        self.position = position
        self.multiplicity = 1
        self.next_to_add = 0  # Will keep count when reordering elements

    def compute_all_combinations(self, ncells):
        """
        Compute all combinations (without considering order) of the elements
        in the group.

        Parameters
        ----------
        ncells : int
            Size of supercell
        """
        self.combinations = []
        natoms = self.multiplicity * ncells
        for combination in \
                itertools.combinations_with_replacement(self.iter_species,
                                                        natoms):
            self.combinations.append(combination)



"""
This module provides the Structure class.
"""

from ase import Atoms

from _icet import Structure


@classmethod
def _structure_from_atoms(self, conf: Atoms):
    """
    Returns the input configuration as an icet Structure object.

    Parameters
    ----------
    conf
        input configuration

    Returns
    -------
    atomic configuration
    """
    return self(conf.positions,
                conf.get_chemical_symbols(),
                conf.cell,
                conf.pbc.tolist())


Structure.from_atoms = _structure_from_atoms


def _structure_to_atoms(self) -> Atoms:
    """
    Returns the structure as an ASE Atoms object.

    Returns
    -------
    atomic configuration
    """
    import ase
    conf = ase.Atoms(pbc=self.pbc)
    conf.set_cell(self.cell)
    for symbol, position in zip(self.chemical_symbols, self.positions):
        conf.append(ase.Atom(symbol, position))
    conf.set_positions(self.get_positions())
    conf.set_chemical_symbols(self.get_chemical_symbols())
    return conf


Structure.to_atoms = _structure_to_atoms


def _repr_function(self) -> str:
    s = ['Cell:']
    s += ['{}\n'.format(self.cell)]
    s += ['Element and positions:']
    for symbol, position in zip(self.chemical_symbols, self.positions):
        s += ['{}  [{:}  {:}  {:}]'.format(symbol, *position)]
    return '\n'.join(s)


Structure.__repr__ = _repr_function



from typing import Dict, List

import numpy as np
from math import isclose

from ase import Atoms
from ase.units import kB
from ase.data import atomic_numbers, chemical_symbols

from .. import DataContainer
from ..calculators.base_calculator import BaseCalculator
from .thermodynamic_base_ensemble import ThermodynamicBaseEnsemble
from .vcsgc_ensemble import get_phis
from .semi_grand_canonical_ensemble import get_chemical_potentials


class HybridEnsemble(ThermodynamicBaseEnsemble):
    """

    Instances of this class allows one to combine multiple ensembles.
    In particular, a dictionary should be provided for each ensemble,
    which must include the type (`ensemble`) as well as the index of
    the sublattice (`sublattice_index`). In addition, it is possible
    to provide a list of allowed symbols (`allowed_symbols`), which
    must represent a subset of the elements that can occupy the sites
    on the specified sublattice. Note that additional arguments are
    required for the SGC and VCSGC ensembles, namely chemical
    potentials (`chemical_potentials`) for the former and constraint
    parameters (`phis` and `kappa`) for the latter. For more detailed
    information regarding the different ensembles, please see
    :class:`CanonicalEnsemble <mchammer.ensembles.CanonicalEnsemble>`,
    :class:`SemiGrandCanonicalEnsemble
    <mchammer.ensembles.SemiGrandCanonicalEnsemble>`, and
    :class:`VCSGCEnsemble <mchammer.ensembles.VCSGCEnsemble>`.

    This class is particularly useful for effectively sampling complex
    multi-component systems with several active sublattices, in which
    case different ensembles can be defined for each of the latter.
    The fact that it is possible to set the allowed chemical symbols
    means that one can vary the concentrations of a few selected
    species, with the help of a VCSGC or semi-grand canonical
    ensemble, while still allowing swaps between any two sites, using
    a canonical ensemble (see also the below example). It is advisable
    to carefully consider how to define the ensemble probabilities. By
    default the ensembles are weighted by the sizes of the
    corresponding sublattices, which should give suitable
    probabilities in most cases. As is shown in the example below, it
    might be prudent to provide different values if allowed symbols
    are provided as well as for cases where there are several
    ensembles that are active on different sublattices.

    Parameters
    ----------

    structure : :class:`Atoms <ase.Atoms>`
        atomic configuration to be used in the Monte Carlo simulation;
        also defines the initial occupation vector
    calculator : :class:`BaseCalculator <mchammer.calculators.ClusterExpansionCalculator>`
        calculator to be used for calculating the potential changes
        that enter the evaluation of the Metropolis criterion
    temperature : float
        temperature :math:`T` in appropriate units [commonly Kelvin]
    ensemble_specs: List[Dict]
        A list of dictionaries, which should contain the following items:

            * 'ensemble', which could be either "vcsgc"; "semigrand"
              or "canonical", lowercase and uppercase letters or any
              combination thereof are accepted (required)
            * 'sublattice_index', index for the sublattice of
              interest (required)
            * 'allowed_symbols', list of allowed chemical symbols
              (default: read from ClusterSpace)
            * 'chemical_potentials', a dictionary of chemical
              potentials for each species
              :math:`\\mu_i`; the key denotes the species, the value
              :specifies the chemical potential in units that are
              :consistent with the underlying cluster expansion (only
              :applicable and required for SGC ensembles)
            * 'phis ', dictionary with average constraint parameters
              ':math:`\\phi_i`; the key denotes the species; for a
              N-component sublattice, there should be N - 1
              different `\\phi_i` (referred to as
              :math:`\\bar{\\phi}` in [SadErh12]_; only applicable
              and required for VCSGC ensembles, see also
              :class:`VCSGCEnsemble <mchammer.ensembles.VCSGCEnsemble>`)
            * 'kappa', parameter that constrains the variance of the
              'concentration (referred to as
              :math:`\\bar{\\kappa}` in [SadErh12]_; only applicable
              :and required for VCSGC ensembles)

    probabilities: List[float]
        list of floats with the probabilities for choosing a
        particular ensemble with the same length as ensemble specs.
        If left unspecified the probabilties are weighted by the
        sizes of the associated sublattices
    boltzmann_constant : float
        Boltzmann constant :math:`k_B` in appropriate units, i.e.
        units that are consistent with the underlying cluster
        expansion and the temperature units [default: eV/K]
    user_tag : str
        human-readable tag for ensemble [default: None]
    random_seed : int
        seed for the random number generator used in the Monte Carlo
        simulation
    data_container : str
        name of file the data container associated with the
        ensemble will be written to; if the file
        exists it will be read, the data container will be appended,
        and the file will be updated/overwritten
    data_container_write_period : float
        period in units of seconds at which the data container is
        written to file; writing periodically to file provides both
        a way to examine the progress of the simulation and to
        back up the data [default: np.inf]
    ensemble_data_write_interval : int
        interval at which data is written to the data container;
        this includes for example the current value of the
        calculator (i.e. usually the energy) as well as ensembles
        specific fields such as temperature or the number of atoms
        of different species
    trajectory_write_interval : int
        interval at which the current occupation vector of the
        atomic configuration is written to the data container.

    Example
    -------
    The following snippet illustrates how to carry out a simple Monte Carlo
    simulation using a combination of one canonical and one VCSGC ensemble.
    Specifically, the concentration of one species (Au) is kept constant
    while the others (Ag and Pd) are varied, while swaps are still allowed.
    Here, the parameters of the cluster expansion are set to emulate a simple
    Ising model in order to obtain an example that can be run without
    modification. In practice, one should of course use a proper cluster
    expansion::

        from ase.build import bulk
        from icet import ClusterExpansion, ClusterSpace
        from mchammer.calculators import ClusterExpansionCalculator
        from mchammer.ensembles import HybridEnsemble

        # prepare cluster expansion
        # the setup emulates a second nearest-neighbor (NN) Ising model
        # (zerolet and singlet ECIs are zero; only first and second neighbor
        # pairs are included)
        prim = bulk('Au')
        cs = ClusterSpace(prim, cutoffs=[4.3],
                          chemical_symbols=['Ag', 'Au', 'Pd'])
        ce = ClusterExpansion(
            cs, [0, 0, 0, 0.1, 0.1, 0.1, -0.02, -0.02, -0.02])

        # define structure object
        structure = prim.repeat(3)
        for i, atom in enumerate(structure):
            if i % 2 == 0:
                atom.symbol = 'Ag'
            elif i % 3 == 0:
                atom.symbol = 'Pd'

        # the default probabilities for this case would be [0.5, 0.5], but
        # since the VCSGC ensemble only performs flips on a subset of all
        # sites on the sublattice, namely those originally occupied by Ag
        # and Pd atoms, specific values will be provided
        weights = [len(structure),
                   len([s for s in structure.get_chemical_symbols() if s != 'Au'])]
        norm = sum(weights)
        probabilities = [w / norm for w in weights]

        # set up and run MC simulation
        calc = ClusterExpansionCalculator(structure, ce)
        ensemble_specs = [
            {'ensemble': 'canonical', 'sublattice_index': 0},
            {'ensemble': 'vcsgc', 'sublattice_index': 0,
             'phis': {'Ag': -0.2}, 'kappa': 200,
             'allowed_symbols':['Ag', 'Pd']}]
        mc = HybridEnsemble(structure=structure, calculator=calc,
                            ensemble_specs=ensemble_specs,
                            temperature=600, probabilities=probabilities,
                            data_container='myrun_hybrid.dc')
        mc.run(100)  # carry out 100 trial steps
    """

    def __init__(self,
                 structure: Atoms,
                 calculator: BaseCalculator,
                 temperature: float,
                 ensemble_specs: List[Dict],
                 probabilities: List[float] = None,
                 boltzmann_constant: float = kB,
                 user_tag: str = None,
                 random_seed: int = None,
                 data_container: str = None,
                 data_container_write_period: float = np.inf,
                 ensemble_data_write_interval: int = None,
                 trajectory_write_interval: int = None) -> None:

        # define available ensembles
        self._ensemble_trial_steps = dict([
            ('canonical', self.do_canonical_swap),
            ('semi-grand', self.do_sgc_flip),
            ('vcsgc', self.do_vcsgc_flip),
            ])

        self._ensemble_parameters = dict(temperature=temperature)

        self._trial_steps_per_ensemble = {"ensemble_{}".format(i): 0 for i in
                                          range(len(ensemble_specs))}

        # process the list of ensembles and parameters
        self._process_ensemble_specs(ensemble_specs)

        super().__init__(
            structure=structure,
            calculator=calculator,
            user_tag=user_tag,
            random_seed=random_seed,
            boltzmann_constant=boltzmann_constant,
            data_container=data_container,
            data_container_class=DataContainer,
            data_container_write_period=data_container_write_period,
            ensemble_data_write_interval=ensemble_data_write_interval,
            trajectory_write_interval=trajectory_write_interval)

        # postprocess the list of ensembles and parameters
        self._postprocess_ensemble_args()

        # set the probabilities
        self._process_probabilities(probabilities)

    @property
    def temperature(self) -> float:
        """ Current temperature """
        return self._ensemble_parameters['temperature']

    @property
    def probabilities(self) -> float:
        """ Ensemble propabilities """
        return self._probabilities

    @property
    def trial_steps_per_ensemble(self) -> Dict[str, float]:
        """ Number of Monte Carlo trial steps for each ensemble """
        return self._trial_steps_per_ensemble

    def _process_ensemble_specs(
            self, ensemble_specs: List[Dict]):
        """Process the list of ensembles and parameters

        Parameters
        ----------
         ensemble_specs: List[Dict]
            A list of dictionaries, which should contain the following items:
            * 'ensemble', which could be either "vcsgc"; "semigrand" or "canonical", lowercase and
            upercase letters or any combination thereof are accepted
            * 'sublattice_index', index for the sublattice of interest
            * 'allowed_symbols', list of allowed chemical symbols
            * 'chemical_potentials', a dictionary of chemical potentials for each species
            :math:`\\mu_i`; the key denotes the species, the value specifies the chemical potential
            in units that are consistent with the underlying cluster expansion
            * 'phis ', dictionary with average constraint parameters :math:`\\phi_i`; the key
            denotes the species; for a N-component sublattice, there should be N - 1 different
            `\\phi_i`
            * 'kappa', parameter that constrains the variance of the concentration
        """

        ensemble_args = []

        for ind, ensemble_spec in enumerate(ensemble_specs):

            ensemble_arg = {}
            tag = "ensemble_{}".format(ind)
            ensemble_arg['tag'] = tag

            # check the ensemble name
            if 'ensemble' not in ensemble_spec:
                raise ValueError("The dictionary {} lacks the required key"
                                 " 'ensemble'".format(ensemble_spec))
            ensemble = ensemble_spec['ensemble'].lower()
            if ensemble not in self._ensemble_trial_steps.keys():
                msg = ['Ensemble not available']
                msg += ['Please choose one of the following:']
                for key in self._ensemble_trial_steps.keys():
                    msg += [' * ' + key]
                raise ValueError('\n'.join(msg))
            ensemble_arg['ensemble'] = ensemble
            self._ensemble_parameters[tag] = ensemble

            # check that all required keys, and no unknown keys, are present
            keys = ['ensemble', 'sublattice_index', 'allowed_symbols']
            if ensemble == 'semi-grand':
                keys = ['chemical_potentials'] + keys
            elif ensemble == 'vcsgc':
                keys = ['phis', 'kappa'] + keys
            for key in keys[:-1]:
                if key not in ensemble_spec:
                    raise ValueError("The dictionary {} lacks the key '{}', which is required for"
                                     " {} ensembles".format(ensemble_spec, key, ensemble))
            for key in ensemble_spec.keys():
                if key not in keys:
                    raise ValueError("Unknown key '{}', for a {} ensemble,"
                                     " in the dictionary {}".format(key, ensemble, ensemble_spec))

            # record the sublattice index
            ensemble_arg['sublattice_index'] = ensemble_spec['sublattice_index']

            # process chemical potentials
            if 'chemical_potentials' in ensemble_spec:
                chemical_potentials = get_chemical_potentials(ensemble_spec['chemical_potentials'])
                ensemble_arg['chemical_potentials'] = chemical_potentials
                for atnum, chempot in chemical_potentials.items():
                    mu_sym = '{}_mu_{}'.format(tag, chemical_symbols[atnum])
                    self._ensemble_parameters[mu_sym] = chempot

            # process phis
            if 'phis' in ensemble_spec:
                phis = get_phis(ensemble_spec['phis'])
                ensemble_arg['phis'] = phis
                for sym, phi in phis.items():
                    if isinstance(sym, str):
                        chemical_symbol = sym
                    else:
                        chemical_symbol = chemical_symbols[sym]
                    phi_sym = '{}_phi_{}'.format(tag, chemical_symbol)
                    self._ensemble_parameters[phi_sym] = phi

            # process kappa
            if 'kappa' in ensemble_spec:
                ensemble_arg['kappa'] = ensemble_spec['kappa']
                self._ensemble_parameters['{}_kappa'.format(tag)] = ensemble_spec['kappa']

            # record the allowed chemical symbols
            if 'allowed_symbols' in ensemble_spec:
                ensemble_arg['allowed_symbols'] = ensemble_spec['allowed_symbols']

            ensemble_args.append(ensemble_arg)

        self._ensemble_args = ensemble_args

    def _postprocess_ensemble_args(self):
        """Process the list of dictionaries with ensemble specific parameters
        """

        for i in range(len(self._ensemble_args)):

            # check the sublattice index
            self._check_sublattice_index(self._ensemble_args[i]['sublattice_index'])

            # extract the allowed species
            if 'allowed_symbols' in self._ensemble_args[i]:
                self._ensemble_args[i]['allowed_species'] =\
                    self._extract_allowed_species(self._ensemble_args[i]['allowed_symbols'],
                                                  self._ensemble_args[i]['sublattice_index'])
                del self._ensemble_args[i]['allowed_symbols']
            else:
                self._ensemble_args[i]['allowed_species'] = None

            if self._ensemble_args[i]['ensemble'] == 'vcsgc':
                # Check that each sublattice has N - 1 phis
                count_specified_elements = 0
                if self._ensemble_args[i]['allowed_species'] is None:
                    allowed_species =\
                        self.sublattices[self._ensemble_args[i]['sublattice_index']].atomic_numbers
                else:
                    allowed_species = self._ensemble_args[i]['allowed_species']
                for number in allowed_species:
                    if number in self._ensemble_args[i]['phis'].keys():
                        count_specified_elements += 1
                if count_specified_elements != len(allowed_species) - 1:
                    raise ValueError("phis must be set for N - 1 elements on a sublattice with"
                                     " N elements")

    def _check_sublattice_index(self, sublattice_index: int):
        """Check the 'sublattice_index' item in the 'ensemble_spec' dictionary

        Parameters
        ----------
        sublattice_index:
            Specific sublattice to consider provided as as an index or a symbol
        """

        if not isinstance(sublattice_index, int):
            raise TypeError("'sublattice_index' must be an integer, not"
                            " {}".format(type(sublattice_index)))

        # check that the sublattice exists
        if sublattice_index not in range(len(self.sublattices)):
            raise ValueError("There is no sublattice with index {}".format(sublattice_index))

        # check that the sublattice is active
        if len(self.sublattices[sublattice_index].chemical_symbols) == 1:
            raise ValueError("The sublattice {} is inactive".format(sublattice_index))

    def _extract_allowed_species(self, allowed_symbols: List[str], sublattice_index: int
                                 ) -> List[int]:
        """Check and extract the allowed species from the 'allowed_symbols' in the 'ensemble_spec'
        dictionary

        Parameters
        ----------
        allowed_symbols:
            list of allowed chemical symbols
        sublattice_index:
            Index for the relevant sublattice
        """

        if not isinstance(allowed_symbols, list) or not all(
                [isinstance(i, str) for i in allowed_symbols]):
            raise TypeError(
                "'allowed_symbols' must be a List[str], not {}".format(type(allowed_symbols)))
        for symbol in allowed_symbols:
            if symbol not in self.sublattices[sublattice_index].chemical_symbols:
                raise ValueError("The species {} is not allowed on sublattice"
                                 " {}".format(symbol, sublattice_index))

        return [atomic_numbers[s] for s in allowed_symbols]

    def _process_probabilities(self, probabilities: List[float]):
        """Process the list of probabilities

        Parameters
        ----------
        probabilities:
            list of floats with the probabilities for choosing a particular ensemble with the same
            length as self._ensemble_args.
        """

        if probabilities is None:
            # use the sizes of the associated sublattices when calculating the ensemble
            # probabilities
            weights = [len(self.sublattices[ensemble_arg['sublattice_index']].indices) for
                       ensemble_arg in self._ensemble_args]
            norm = sum(weights)
            probabilities = [w / norm for w in weights]
        else:
            if len(probabilities) != len(self._ensemble_args):
                raise ValueError("The number of probabilities must be match the number of"
                                 " ensembles")

            if not isclose(sum(probabilities), 1.0):
                raise ValueError("The sum of all probabilities must be equal to 1")

        self._probabilities = probabilities

    def _do_trial_step(self):
        """ Carries out one Monte Carlo trial step. """

        # randomly pick an ensemble
        ensemble_arg = np.random.choice(self._ensemble_args, p=self._probabilities)

        # count number of trial steps for each ensemble
        self._trial_steps_per_ensemble[ensemble_arg['tag']] += 1

        if ensemble_arg['ensemble'] == 'canonical' and not self.configuration.is_swap_possible(
                ensemble_arg['sublattice_index'], ensemble_arg['allowed_species']):
            return 0
        else:
            arguments = {key: val for key, val in ensemble_arg.items() if key not in
                         ['ensemble', 'tag']}
            return self._ensemble_trial_steps[ensemble_arg['ensemble']](**arguments)

    def _get_ensemble_data(self) -> Dict:
        """
        Returns a dict with the default data of the ensemble. This includes
        atom counts and free energy derivative.
        """
        data = super()._get_ensemble_data()

        ensemble_types = [e['ensemble'] for e in self._ensemble_args]

        # free energy derivative
        if 'vcsgc' in ensemble_types:
            for ensemble_arg in self._ensemble_args:
                if 'vcsgc' == ensemble_arg['ensemble']:
                    data.update(self._get_vcsgc_free_energy_derivatives(
                        ensemble_arg['phis'], ensemble_arg['kappa'],
                        ensemble_arg['sublattice_index']))

        # species counts
        if any([e in ensemble_types for e in ['vcsgc', 'semi-grand']]):
            data.update(self._get_species_counts())

        return data



"""
Definition of the base observer class.
"""

from abc import ABC, abstractmethod


class BaseObserver(ABC):
    """
    Base observer class.

    Parameters
    ----------
    interval : int
        the observation interval
    tag : str
        human readable tag used for identifying the observer

    Attributes
    ----------
    tag : str
        human readable tag used for identifying the observer
    interval : int
        the observation interval
    """

    def __init__(self, return_type: type, interval: int = None,
                 tag: str = 'BaseObserver') -> None:
        self.tag = tag
        self.interval = interval
        self._return_type = return_type

    @property
    def return_type(self) -> type:
        """Data type of the observed data."""
        return self._return_type

    @abstractmethod
    def get_observable(self):
        """
        Method used for extracting data.

        Returns
        -------
        self.return_type()

        When implementing this method use the following names for the
        following types of data:

        ASE Atoms object : `structure`.
        list of chemical species : `species`.
        icet cluster expansion : `cluster_expansion`.
        mchammer calculator : `calculator`.
        """
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import sys
from glob import glob

import setuptools
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext


icet_cpp_module = Extension(
    '_icet',
    glob('src/*.cpp'),
    include_dirs=[
        'src/3rdparty/pybind11/include/',
        'src/3rdparty/boost_1_68_0/',
        'src/3rdparty/eigen3/'
    ],
    language='c++')


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """
    Returns a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """
    Returns the -std=c++[11/14] compiler flag.

    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if sys.platform == 'darwin':
            if has_flag(self.compiler, '-stdlib=libc++'):
                opts.append('-stdlib=libc++')
        if ct == 'unix':
            opts.append("-DVERSION_INFO='{}'"
                        .format(self.distribution.get_version()))
            opts.append(cpp_flag(self.compiler))
            opts.append('-O3')
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
        elif ct == 'msvc':
            opts.append("/DVERSION_INFO=\\'{}\\'"
                        .format(self.distribution.get_version()))
        for ext in self.extensions:
            ext.extra_compile_args = opts
            ext.extra_link_args = opts
        build_ext.build_extensions(self)


if sys.version_info < (3, 5, 0, 'final', 0):
    raise SystemExit('Python 3.5 or later is required!')


with open('README.rst', encoding='utf-8') as fd:
    long_description = fd.read()

with open('icet/__init__.py', encoding='utf-8') as fd:
    try:
        lines = ''
        for item in fd.readlines():
            item = item
            lines += item + '\n'
    except Exception as exc:
        raise Exception('Caught exception {}'.format(exc))


author_list = ['Mattias Ångqvist',
               'William A. Muñoz',
               'J. Magnus Rahm',
               'Erik Fransson',
               'Céline Durniak',
               'Piotr Rozyczko',
               'Thomas Holm Rod',
               'Paul Erhart']
authors = ', '.join(author_list[:-1]) + ', and ' + author_list[-1]

version = re.search("__version__ = '(.*)'", lines).group(1)
maintainer = re.search("__maintainer__ = '(.*)'", lines).group(1)
email = re.search("__email__ = '(.*)'", lines).group(1)
description = re.search("__description__ = '(.*)'", lines).group(1)
url = re.search("__url__ = '(.*)'", lines).group(1)
license = re.search("__license__ = '(.*)'", lines).group(1)

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: {}'.format(license),
    'Topic :: Scientific/Engineering :: Physics']


if __name__ == '__main__':

    setup(
        name='icet',
        version=version,
        author=authors,
        author_email=email,
        description=description,
        long_description=long_description,
        ext_modules=[icet_cpp_module],
        install_requires=['pybind11>=2.2',
                          'ase',
                          'numpy',
                          'scipy',
                          'sklearn',
                          'pandas>=0.23',
                          'spglib>1.12.0'],
        packages=find_packages(),
        cmdclass={'build_ext': BuildExt},
        zip_safe=False,
        classifiers=classifiers,
        license=license,
        url=url,
    )



"""
This module provides the ClusterCounts class.
"""

from collections import OrderedDict
from _icet import ClusterCounts as _ClusterCounts
from _icet import Cluster
from ase import Atoms
from icet.core.orbit_list import OrbitList
from icet import Structure
from .local_orbit_list_generator import LocalOrbitListGenerator


class ClusterCounts(_ClusterCounts):
    """
    Provides an interface for inspecting cluster counts.

    Parameters
    ----------
    orbit_list : icet.OrbitList
        orbit list for a primitive structure
    structure : ase.Atoms
        supercell of the structure that `orbit_list` is based on

    Attributes
    ----------
    cluster_counts : OrderedDict
        keys are representative clusters (icet Cluster objects) for all
        distinct orbits in the orbit list, values are dicts where the key is
        the chemical symbols of the species in a cluster, and the value is
        the number of counts of such clusters, e.g.,
        {('Au', 'Ag'): 3, ('Au', 'Au'): 5}.
    """

    def __init__(self, orbit_list: OrbitList, structure: Atoms):
        self._orbit_list = orbit_list
        self._structure = Structure.from_atoms(structure)
        # call (base) C++ constructor
        _ClusterCounts.__init__(self)
        self.cluster_counts = self._count_clusters()

    def _count_clusters(self, keep_order_intact=False, permute_sites=True):
        """
        Counts all clusters in a structure by finding their local orbit list.

        Parameters
        ----------
        keep_order_intact : bool
            if true the order in the cluster will be sorted
        permute_sites : bool
            if true will permute the sites so they are in the
            symmetrically equivalent order as the representative sites
        """

        local_orbit_list_generator = LocalOrbitListGenerator(
            self._orbit_list, self._structure)

        for i in range(
                local_orbit_list_generator.get_number_of_unique_offsets()):
            self.count_orbit_list(
                self._structure,
                local_orbit_list_generator.generate_local_orbit_list(i),
                keep_order_intact, permute_sites)

        sorted_cluster_counts = \
            OrderedDict(sorted(self.get_cluster_counts().items()))
        return sorted_cluster_counts

    def __str__(self):
        """
        String representation of ClusterCounts object that provides an
        overview of the clusters (cluster, chemical symbol, and count).
        """
        tuplets = {1: 'Singlet', 2: 'Pair', 3: 'Triplet',
                   4: 'Quadruplet', 5: 'Pentuplet'}
        width = 60
        s = ['{s:=^{n}}'.format(s=' Cluster Counts ', n=width)]

        first = True
        for cluster, counts in self.cluster_counts.items():
            if not first:
                s += ['']
            else:
                first = False

            # Add a description of the orbit to the string
            tuplet_type = tuplets.get(len(cluster.sites),
                                      '{}-tuplet'.format(len(cluster.sites)))
            s += ['{}: {} {:} {:.4f}'
                  .format(tuplet_type, cluster.sites,
                          cluster.distances, cluster.radius)]

            # Print the actual counts together with the species they refer to
            for chemical_symbols, count in counts.items():
                t = ['{:3} '.format(el) for el in chemical_symbols]
                s += ['{} {}'.format(''.join(t), count)]
        s += [''.center(width, '=')]
        return '\n'.join(s)

    def __getitem__(self, key):
        """
        Returns cluster count (Cluster object and dict with counts) for a
        ClusterCounts object.

        Parameters
        ----------
        key : int or icet.Cluster
            if int, return the key-th counts;
            if Cluster, return the counts belonging to that cluster
        """
        if isinstance(key, int):
            return list(self.cluster_counts.values())[key]
        elif isinstance(key, Cluster):
            return self.cluster_counts[key]
